# Pytorch
import torch
from torch import nn, optim
from torch.nn import functional as F
from torch.utils.data import DataLoader
from torch.utils.data.sampler import SubsetRandomSampler
from torch.distributions import kl_divergence as kld
from torch.optim.lr_scheduler import ReduceLROnPlateau


# Third Party
import numpy as np
from pathlib import Path
import pandas as pd
import scanpy as sc

# Built-in
import time
import numpy as np
from scipy.sparse import issparse
from sklearn.utils import class_weight
from collections import Counter
from itertools import chain
from copy import deepcopy
import json
from typing import Mapping, Union, Iterable, Tuple, Optional, Mapping, Dict
import os
import warnings


# Package
from ._primitives import *
from ..utils._tensor_utils import one_hot, get_k_elements, get_last_k_elements, get_elements
from ..utils._decorators import typed
from ..utils._loss import LossFunction
from ..utils._logger import mt, mw, Colors, get_tqdm

from ..utils._utilities import random_subset_by_key_fast, euclidean
from ..utils._compat import Literal
from ..tools._umap import transfer_umap
from ..utils._utilities import get_default_device
from ..preprocessing._preprocess import subset_adata_by_genes_fill_zeros

MODULE_PATH = Path(__file__).parent
warnings.filterwarnings("ignore")


class scAtlasVAE(ReparameterizeLayerBase, MMDLayerBase):
    """
    VAE model for atlas-level integration and label transfer

    :param adata: AnnData. If provided, initialize the model with the adata.
    :param hidden_stacks: List[int]. Number of hidden units in each layer. Default: [128] (one hidden layer with 128 units)
    :param n_latent: int. Number of latent dimensions. Default: 10
    :param n_batch: int. Number of batch. Default: 0
    :param n_label: int. Number of label. Default: 0
    :param n_additional_batch: Optional[Iterable[int]]. Number of categorical covariate. Default: None
    :param batch_key: str. Batch key in adata.obs. Default: None
    :param label_key: str. Label key in adata.obs. Default: None
    :param dispersion: Literal["gene", "gene-batch", "gene-cell"]. Dispersion method. Default: "gene-cell"
    :param log_variational: bool. If True, log the variational distribution. Default: True
    :param total_variational: bool. If True, normalize the counts with library size. Default: False
    :param bias: bool. If True, use bias in the linear layer. Default: True
    :param use_batch_norm: bool. If True, use batch normalization. Default: True
    :param use_layer_norm: bool. If True, use layer normalization. Default: False
    :param batch_hidden_dim: int. Number of hidden units in the batch embedding layer. Default: 8
    :param batch_embedding: Literal["embedding", "onehot"]. Batch embedding method. Default: "batch_embedding"
    :param constrain_latent_method: Literal['mse', 'normal']. Method to constrain the latent embedding. Default: 'mse'
    :param constrain_latent_embedding: bool. If True, constrain the latent embedding. Default: False
    :param constrain_latent_key: str. Key to the data to constrain the latent embedding. Default: 'X_gex'
    :param encode_libsize: bool. If True, encode the library size. Default: False
    :param decode_libsize: bool. If True, decode the library size. Default: True
    :param dropout_rate: float. Dropout rate. Default: 0.1
    :param activation_fn: nn.Module. Activation function. Default: nn.ReLU
    :param inject_batch: bool. If True, inject batch information. Default: True
    :param inject_label: bool. If True, inject label information. Default: False
    :param inject_additional_batch: bool. If True, inject categorical covariate information. Default: True
    :param unlabel_key: str. key for unlabeled cells. Default: "undefined"
    :param mmd_key: Optional[Literal['batch']]. If provided, use MMD loss. Default: None (do not use MMD loss)
    :param pretrained_state_dict: torch.device or str. Build the model loading the pretrained state dict
    :param device: Optional[Union[str, torch.device]]. Device to use. Default: determined by availablility of CUDA device

    :example:
        >>> import scatlasvae
        >>> model = scatlasvae.model.scAtlasVAE(
        >>>    adata,
        >>>    batch_key = ['sample_name','study_name'],
        >>>    label_key = ['cell_type', 'cell_subtype'],
        >>> )
    """
    def __init__(self, *,
       adata: Optional[sc.AnnData] = None,
       hidden_stacks: List[int] = [128],
       n_latent: int = 10,
       n_batch: int = 0,
       n_label: int = 0,
       n_additional_batch: Optional[Iterable[int]] = None,
       n_additional_label: Optional[Iterable[int]] = None,
       batch_key = None,
       label_key = None,
       dispersion:  Literal["gene", "gene-batch", "gene-cell"] = "gene-cell",
       log_variational: bool = True,
       total_variational: bool = False,
       bias: bool = True,
       use_batch_norm: bool = True,
       use_layer_norm: bool = False,
       batch_hidden_dim: int = 8,
       batch_embedding: Literal["embedding", "onehot"] = "embedding",
       reconstruction_method: Literal['mse', 'zg', 'zinb'] = 'zinb',
       constrain_n_label: bool = True,
       constrain_n_batch: bool = True,
       constrain_latent_method: Literal['mse', 'normal'] = 'mse',
       constrain_latent_embedding: bool = False,
       constrain_latent_key: str = 'X_gex',
       encode_libsize: bool = False,
       decode_libsize: bool = True,
       dropout_rate: float = 0.1,
       activation_fn: nn.Module = nn.ReLU,
       inject_batch: bool = True,
       inject_label: bool = False,
       inject_additional_batch: bool = True,
       mmd_key: Optional[Literal['batch','additional_batch','both']] = None,
       unlabel_key: str = 'undefined',
       device: Optional[Union[str, torch.device]] = None,
       pretrained_state_dict: Optional[Mapping[str, torch.Tensor]] = None,
    ) -> None:
        if device is None:
            device = get_default_device()
        
        super(scAtlasVAE, self).__init__()
        if adata.X.dtype != np.int32 and reconstruction_method in ['zinb', 'nb']:
            mw("adata.X is not of type np.int32. \n" + \
                " "*40 + "\tCheck whether you are using raw count matrix.")
            # adata.X = adata.X.astype(np.int32)

        self.adata = adata
        self.in_dim = adata.shape[1] if adata else -1
        self.n_hidden = hidden_stacks[-1]
        self.n_latent = n_latent
        self.n_additional_batch = n_additional_batch
        self.n_additional_label = n_additional_label
        self._hidden_stacks = hidden_stacks

        if n_batch > 0 and not batch_key:
            raise ValueError("Please provide a batch key if n_batch is greater than 0")
        if n_label > 0 and not label_key:
            raise ValueError("Please provide a label key if n_batch is greater than 0")

        self.label_key = label_key if isinstance(label_key, str) else label_key[0] if label_key is not None and isinstance(label_key, Iterable) else None
        self.label_category = None 
        self.label_category_summary = None 
        self.batch_key = batch_key if isinstance(batch_key, str) else batch_key[0] if batch_key is not None and isinstance(batch_key, Iterable) else None
        self.batch_category = None 
        self.batch_category_summary = None 
        self.additional_batch_keys = None if isinstance(batch_key, str) or (isinstance(batch_key, Iterable) and len(batch_key) == 1) else batch_key[1:] if batch_key is not None else None
        self.additional_batch_category = None 
        self.additional_batch_category_summary = None 
        self.additional_label_keys = None if isinstance(label_key, str) or (isinstance(label_key, Iterable) and len(label_key) == 1) else label_key[1:] if label_key is not None else None
        self.additional_label_category = None 
        self.additional_label_category_summary = None 


        self.n_batch = n_batch
        self.n_label = n_label
        self.unlabel_key = unlabel_key
        self.new_adata_code = None
        self.log_variational = log_variational
        self.total_variational = total_variational
        self.mmd_key = mmd_key
        self.reconstruction_method = reconstruction_method
        self.constrain_latent_embedding = constrain_latent_embedding
        self.constrain_latent_method = constrain_latent_method
        self.constrain_latent_key = constrain_latent_key
        self.constrain_n_label = constrain_n_label
        self.constrain_n_batch = constrain_n_batch

        self.device=device

        self.initialize_dataset()

        self.batch_embedding = batch_embedding
        if batch_embedding == "onehot":
            batch_hidden_dim = self.n_batch
        self.batch_hidden_dim = batch_hidden_dim
        self.inject_batch = inject_batch
        self.inject_label = inject_label
        self.inject_additional_batch = inject_additional_batch
        self.encode_libsize = encode_libsize
        self.decode_libsize = decode_libsize
        self.dispersion = dispersion


        self.fcargs = dict(
            bias           = bias,
            dropout_rate   = dropout_rate,
            use_batch_norm = use_batch_norm,
            use_layer_norm = use_layer_norm,
            activation_fn  = activation_fn,
            device         = device
        )



        #############################
        # Model Trainable Variables #
        #############################

        if self.dispersion == "gene":
            self.px_rate = torch.nn.Parameter(torch.randn(self.in_dim))
        elif self.dispersion == "gene-batch":
            self.px_rate = torch.nn.Parameter(torch.randn(self.in_dim, self.n_batch))
        elif self.dispersion == "gene-cell":
            pass
        else:
            raise ValueError(
                "dispersion must be one of ['gene', 'gene-batch',"
                " 'gene-cell'], but input was "
                "{}.format(self.dispersion)"
            )

        ############
        # ENCODERS #
        ############

        self.encoder = SAE(
            self.in_dim if not self.encode_libsize else self.in_dim + 1,
            stacks = hidden_stacks,
            # n_cat_list = [self.n_batch] if self.n_batch > 0 else None,
            cat_dim = batch_hidden_dim,
            cat_embedding = batch_embedding,
            encode_only = True,
            **self.fcargs
        )

        # The latent cell representation z ~ Logisticnormal(0, I)
        self.z_mean_fc = nn.Linear(self.n_hidden, self.n_latent)
        self.z_var_fc = nn.Linear(self.n_hidden, self.n_latent)
        self.z_transformation = nn.Softmax(dim=-1)

        ############
        # DECODERS #
        ############

        if self.n_additional_batch_ is not None and self.inject_additional_batch:
            if self.n_batch > 0 and self.n_label > 0 and inject_batch and inject_label:
                decoder_n_cat_list = [self.n_batch, self.n_label, *self.n_additional_batch]
            elif self.n_batch > 0 and inject_batch:
                decoder_n_cat_list = [self.n_batch, *self.n_additional_batch]
            elif self.n_label > 0 and inject_label:
                decoder_n_cat_list = [self.n_label, *self.n_additional_batch]
            else:
                decoder_n_cat_list = None
        else:
            if self.n_batch > 0 and self.n_label > 0 and inject_batch and inject_label:
                decoder_n_cat_list = [self.n_batch, self.n_label]
            elif self.n_batch > 0 and inject_batch:
                decoder_n_cat_list = [self.n_batch]
            elif self.n_label > 0 and inject_label:
                decoder_n_cat_list = [self.n_label]
            else:
                decoder_n_cat_list = None

        self.decoder_n_cat_list = decoder_n_cat_list

        self.decoder = FCLayer(
            in_dim = self.n_latent,
            out_dim = self.n_hidden,
            n_cat_list = decoder_n_cat_list,
            cat_dim = batch_hidden_dim,
            cat_embedding = batch_embedding,
            use_layer_norm=False,
            use_batch_norm=True,
            dropout_rate=0,
            device=device
        )

        self.px_rna_rate_decoder = nn.Linear(self.n_hidden, self.in_dim)
        self.px_rna_scale_decoder = nn.Sequential(
            nn.Linear(self.n_hidden, self.in_dim),
            nn.Softmax(dim=-1)
        )
        self.px_rna_dropout_decoder = nn.Linear(self.n_hidden, self.in_dim)

        if self.n_label > 0:
            self.fc = nn.Sequential(
                nn.Linear(self.n_latent, self.n_label)
            )

        if self.n_additional_label is not None:
            self.additional_fc = nn.ModuleList([
                nn.Linear(self.n_latent, x) for x in self.n_additional_label
            ])

        self._trained = False

        self.to(device)

        if pretrained_state_dict is not None:
            self.partial_load_state_dict(pretrained_state_dict)


    def __repr__(self):
        return f'{Colors.ORANGE}VAEModel{Colors.NC} object containing:\n' + \
            f'    {Colors.GREEN}adata{Colors.NC}: {self.adata}\n' + \
            f'    {Colors.GREEN}in_dim{Colors.NC}: {Colors.CYAN}{self.in_dim}{Colors.NC}\n' + \
            f'    {Colors.GREEN}n_hidden{Colors.NC}: {Colors.CYAN}{self.n_hidden}{Colors.NC}\n' + \
            f'    {Colors.GREEN}labels{Colors.NC}: {self.label_key} of {Colors.CYAN}{self.n_label}{Colors.NC}\n' if self.batch_key else '' + \
            f'    {Colors.GREEN}batches{Colors.NC}: {self.batch_key} of {Colors.CYAN}{self.n_batch}{Colors.NC}\n' if self.label_key else '' + \
            f'    {Colors.GREEN}additional_batches{Colors.NC}: {self.additional_batch_keys} of {Colors.CYAN}{self.n_additional_batch}{Colors.NC}\n' if self.additional_batch_keys else ''

    def partial_load_state_dict(self, state_dict: Mapping[str, torch.Tensor]):
        """
        Partially load the state dict

        :param state_dict: Mapping[str, torch.Tensor]. State dict to load
        """
        original_state_dict = self.state_dict()
        warned = False
        ignored_keys = []
        for k,v in state_dict.items():
            if k not in original_state_dict.keys():
                mt(f"Warning: {k} not found in the model. Ignoring {k} in the provided state dict.")
                ignored_keys.append(k)
            elif v.shape != original_state_dict[k].shape:
                mw(f"Warning: shape of {k} does not match. \n" + \
                    ' '*40 + "\tOriginal:" + f" {original_state_dict[k].shape},\n" + \
                    ' '*40 + f"\tNew: {v.shape}")
                state_dict[k] = original_state_dict[k]
        for k,v in original_state_dict.items():
            if k not in state_dict.keys():
                mw(f"Warning: {k} not found in the provided state dict. " + \
                     f"Using {k} in the original state dict.")
                state_dict[k] = v
        for i in ignored_keys:
            state_dict.pop(i)
        self.load_state_dict(state_dict)

    def get_config(self):
        """
        Get the model config

        :return: dict. Model config dictionary
        """
        return {
            'hidden_stacks': self._hidden_stacks,
            'n_latent': self.n_latent,
            'n_batch': self.n_batch,
            'n_label': self.n_label,
            'n_additional_batch': self.n_additional_batch,
            'n_additional_label': self.n_additional_label,
            'batch_key': self.batch_key,
            'label_key': self.label_key,
            'additional_batch_keys': self.additional_batch_keys,
            'additional_label_keys': self.additional_label_keys,
            'dispersion': self.dispersion,
            'log_variational': self.log_variational,
            'bias': self.fcargs['bias'],
            'use_batch_norm': self.fcargs['use_batch_norm'],
            'use_layer_norm': self.fcargs['use_layer_norm'],
            'batch_hidden_dim': self.batch_hidden_dim,
            'batch_embedding': self.batch_embedding,
            'reconstruction_method': self.reconstruction_method,
            'encode_libsize': self.encode_libsize,
            'decode_libsize': self.decode_libsize,
            'dropout_rate': self.fcargs['dropout_rate'],
            'activation_fn': self.fcargs['activation_fn'],
            'inject_batch': self.inject_batch,
            'inject_label': self.inject_label,
            'inject_additional_batch': self.inject_additional_batch,
            'mmd_key': self.mmd_key,
            'unlabel_key': self.unlabel_key,
        }

    def save_to_disk(self, path_to_state_dict: Union[str, Path]):
        """
        Save the model to disk

        :param path_to_state_dict: str or Path. Path to save the model
        """
        model_state_dict = self.state_dict()
        model_var_index = self.adata.var.index
        state_dict = {
            "model_state_dict": model_state_dict,
            "model_var_index": model_var_index,
            "model_config": self.get_config(),
            "batch_category": self.batch_category,
            "batch_category_summary": self.batch_category_summary,
            "label_category": self.label_category,
            "label_category_summary": self.label_category_summary,
            "additional_label_category": self.additional_label_category,
            "additional_label_category_summary": self.additional_label_category_summary,
            "additional_batch_category": self.additional_batch_category,
            "additional_batch_category_summary": self.additional_batch_category_summary,
        }
        torch.save(state_dict, path_to_state_dict)

    def load_from_disk(self, path_to_state_dict: Union[str, Path]):
        """
        Load the model from disk

        :param path_to_state_dict: str or Path. Path to load the model
        """
        state_dict = torch.load(path_to_state_dict)
        self.partial_load_state_dict(state_dict["model_state_dict"])

    @staticmethod
    def setup_anndata(
        adata: sc.AnnData, 
        path_to_state_dict: Optional[Union[str, Path]] = None, 
        unlabel_key: str = 'undefined'
    ):
        """
        Setup the model with adata

        :param adata: AnnData. AnnData to setup the model
        :param path_to_state_dict: Optional[str, Path]. Path to the state dict to load
        :param unlabeled_key: str. Default: Undefined

        """
        state_dict = torch.load(path_to_state_dict)
        if 'model_var_index' in state_dict.keys():
            model_var_index = state_dict['model_var_index']
            if any(list(map(lambda x: x not in model_var_index, adata.var.index))) or any(list(map(lambda x: x not in adata.var.index, model_var_index))):
                mw("the provided adata contains variables not in the state dict.")
                mt("         The model will be initialized with the variables in the state dict.")
                adata = subset_adata_by_genes_fill_zeros(adata, list(model_var_index))


        if state_dict["batch_category"] is not None:
            if state_dict['model_config']['batch_key'] not in adata.obs.keys():
                adata.obs[state_dict['model_config']['batch_key']] = unlabel_key
                adata.obs[state_dict['model_config']['batch_key']] = pd.Categorical(
                    list(adata.obs[state_dict['model_config']['batch_key']] ),
                    categories = pd.Categorical(
                        state_dict["batch_category"].categories
                    ).add_categories(unlabel_key).categories
                )
            else:
                adata.obs[state_dict['model_config']['batch_key']] = pd.Categorical(
                    list(adata.obs[state_dict['model_config']['batch_key']] ),
                    categories=state_dict["batch_category"].categories
                )


        if state_dict["label_category"] is not None:
            if state_dict['model_config']['label_key'] not in adata.obs.keys():
                adata.obs[state_dict['model_config']['label_key']] = unlabel_key
                adata.obs[state_dict['model_config']['label_key']] = pd.Categorical(
                    list(adata.obs[state_dict['model_config']['label_key']] ),
                    categories = pd.Categorical(
                        state_dict["label_category"].categories
                    ).add_categories(unlabel_key).categories if \
                    unlabel_key not in state_dict["label_category"].categories else \
                    state_dict["label_category"].categories
                )
            else:
                adata.obs[state_dict['model_config']['label_key']] = pd.Categorical(
                    list(adata.obs[state_dict['model_config']['label_key']] ),
                    categories=state_dict["label_category"].categories
                )

        if state_dict["additional_batch_category"] is not None:
            for i,k in enumerate(state_dict['model_config']['additional_batch_keys']):
                if k not in adata.obs.keys():
                    adata.obs[k] = unlabel_key
                    adata.obs[k] = pd.Categorical(
                        list(adata.obs[k] ),
                        categories = pd.Categorical(
                            state_dict["additional_batch_category"][i].categories
                        ).add_categories(unlabel_key).categories if \
                        unlabel_key not in state_dict["additional_batch_category"][i].categories else \
                        state_dict["additional_batch_category"][i].categories
                    )

                else:
                    adata.obs[k] = pd.Categorical(
                        list(adata.obs[k] ),
                        categories=state_dict["additional_batch_category"][i].categories
                    )

        if state_dict["additional_label_category"] is not None:
            for i,k in enumerate(state_dict['model_config']['additional_label_keys']):
                if k not in adata.obs.keys():
                    adata.obs[k] = unlabel_key
                    adata.obs[k] = pd.Categorical(
                        list(adata.obs[k] ),
                        categories = pd.Categorical(
                            state_dict["additional_label_category"][i].categories
                        ).add_categories(unlabel_key).categories if \
                        unlabel_key not in state_dict["additional_label_category"][i].categories else \
                        state_dict["additional_label_category"][i].categories
                    )
                else:
                    adata.obs[k] = pd.Categorical(
                        list(adata.obs[k] ),
                        categories=state_dict["additional_label_category"][i].categories
                    )

        return adata

    def initialize_dataset(self):
        mt("Initializing dataset into memory")
        if self.batch_key is not None:
            n_batch_ = len(np.unique(self.adata.obs[self.batch_key]))
            if self.n_batch != n_batch_:
                mt(f"warning: the provided n_batch={self.n_batch} does not match the number of batch in the adata.")
                if self.constrain_n_batch:
                    mt(f"         setting n_batch to {n_batch_}")
                    self.n_batch = n_batch_
            if not isinstance(self.adata.obs[self.batch_key], pd.Categorical):
                self.batch_category = pd.Categorical(self.adata.obs[self.batch_key])
            self.batch_category_summary = dict(Counter(self.batch_category))
            for k in self.batch_category.categories:
                if k not in self.batch_category_summary.keys():
                    self.batch_category_summary[k] = 0


        if self.label_key is not None:
            n_label_ = len(np.unique(list(filter(lambda x: x != self.unlabel_key, pd.Categorical(self.adata.obs[self.label_key]).categories))))

            if self.n_label != n_label_:
                mt(f"warning: the provided n_label={self.n_label} does not match the number of batch in the adata.")
                if self.constrain_n_label:
                    mt(f"         setting n_label to {n_label_}")
                    self.n_label = n_label_

            if not isinstance(self.adata.obs[self.label_key], pd.Categorical):
                self.adata.obs[self.label_key] = list(self.adata.obs[self.label_key])
                self.adata.obs[self.label_key] = pd.Categorical(self.adata.obs[self.label_key].fillna(self.unlabel_key))

            self.label_category = pd.Categorical(self.adata.obs[self.label_key])
            self.label_category_summary = dict(Counter(list(filter(lambda x: x != self.unlabel_key, self.label_category))))
            for k in self.label_category.categories:
                if k not in self.label_category_summary.keys() and k != self.unlabel_key:
                    self.label_category_summary[k] = 0

            self.label_category_weight = len(self.label_category) / torch.tensor([
                self.label_category_summary[x] for x in list(filter(lambda x:
                    x != self.unlabel_key,
                    self.label_category.categories
                ))], dtype=torch.float64).to(self.device)

            if self.unlabel_key in self.label_category.categories:
                self.new_adata_code = list(self.label_category.categories).index(self.unlabel_key)

        self.n_additional_label_ = None
        self.additional_label_category = None
        self.additional_label_category_summary = None
        if self.additional_label_keys is not None:
            self.n_cell_additional_label = [len(list(filter(lambda x: x != self.unlabel_key,self.adata.obs[x]))) for x in [self.label_key] + self.additional_label_keys]

            self.n_additional_label_ = [len(np.unique(list(filter(lambda x: x != self.unlabel_key,pd.Categorical(self.adata.obs[x]).categories)))) for x in self.additional_label_keys]

            # self.additional_label_weight = sum(self.n_cell_additional_label) / torch.tensor(self.n_cell_additional_label)
            self.additional_label_weight = torch.tensor([1] * len(self.n_cell_additional_label), dtype=torch.float64).to(self.device)

            if self.n_additional_label == None or len(self.n_additional_label_) != len(self.n_additional_label):
                mt(f"warning: the provided n_additional_label={self.n_additional_label} does not match the number of additional label in the adata.")
                if self.constrain_n_label:
                    mt(f"         setting n_additional_label to {self.n_additional_label_}")
                    self.n_additional_label = self.n_additional_label_
            else:
                for e,(i,j) in enumerate(zip(self.n_additional_label_, self.n_additional_label)):
                    if i != j:
                        mt(f"n_additional_label {self.additional_label_keys[e]} does not match the number in the adata.")
                        if self.constrain_n_label:
                            mt(f"         setting n_additional_label {e} to {i}")
                            self.n_additional_label[e] = i
            self.additional_label_category = [
                pd.Categorical(self.adata.obs[x]) for x in self.additional_label_keys
            ]
            self.additional_label_category_summary = [dict(Counter(x)) for x in self.additional_label_category]
            for i in range(len(self.additional_label_category_summary)):
                for k in self.additional_label_category[i].categories:
                    if k not in self.additional_label_category_summary[i].keys() and k != self.unlabel_key:
                        self.additional_label_category_summary[i][k] = 0

            self.additional_label_category_weight = [len(label_category) / torch.tensor([
                self.additional_label_category_summary[e][x] for x in list(filter(lambda x:
                    x != self.unlabel_key,
                    label_category.categories
                ))], dtype=torch.float64).to(self.device) for e,label_category in enumerate(self.additional_label_category)]



            self.additional_new_adata_code = [list(x.categories).index(self.unlabel_key) if self.unlabel_key in x.categories else -1 for x in self.additional_label_category]

        self.n_additional_batch_ = None
        if self.additional_batch_keys is not None:
            self.n_additional_batch_ = [len(np.unique(self.adata.obs[x])) for x in self.additional_batch_keys]
            if self.n_additional_batch == None or len(self.n_additional_batch_) != len(self.n_additional_batch):
                mt(f"warning: the provided n_additional_batch={self.n_additional_batch} does not match the number of categorical covariate in the adata.")

                if self.constrain_n_batch:
                    mt(f"         setting n_additional_batch to {self.n_additional_batch_}")
                    self.n_additional_batch = self.n_additional_batch_
            else:
                for e,(i,j) in enumerate(zip(self.n_additional_batch_, self.n_additional_batch)):
                    if i != j:
                        mt(f"n_additional_batch {self.additional_batch_keys[e]} does not match the number in the adata.")

                        if self.constrain_n_batch:
                            mt(f"         setting n_additional_batch {e} to {i}")
                            self.n_additional_batch[e] = i
            self.additional_batch_category = [pd.Categorical(self.adata.obs[x]) for x in self.additional_batch_keys]
            self.additional_batch_category_summary = [dict(Counter(x)) for x in self.additional_batch_category]
            for i in range(len(self.additional_batch_category_summary)):
                for k in self.additional_batch_category[i].categories:
                    if k not in self.additional_batch_category_summary[i].keys():
                        self.additional_batch_category_summary[i][k] = 0

        X = self.adata.X

        self._n_record = X.shape[0]
        self._indices = np.array(list(range(self._n_record)))
        batch_categories, label_categories = None, None
        additional_label_categories = None
        additional_batch_categories = None

        if self.batch_key is not None:
            if self.batch_key not in self.adata.obs.columns:
                raise ValueError(f"batch_key {self.batch_key} is not found in AnnData obs")
            batch_categories = np.array(self.batch_category.codes)
        if self.label_key is not None:
            if self.label_key not in self.adata.obs.columns:
                raise ValueError(f"label_key {self.label_key} is not found in AnnData obs")
            label_categories = np.array(self.label_category.codes)
        if self.additional_label_keys is not None:
            for e,i in enumerate(self.additional_label_keys):
                if i not in self.adata.obs.columns:
                    raise ValueError(f"additional_label_keys {i} is not found in AnnData obs")
            additional_label_categories = [np.array(x.codes) for x in self.additional_label_category]
        if self.additional_batch_keys is not None:
            for e,i in enumerate(self.additional_batch_keys):
                if i not in self.adata.obs.columns:
                    raise ValueError(f"additional_batch_keys {i} is not found in AnnData obs")
            additional_batch_categories = [np.array(x.codes) for x in self.additional_batch_category]


        if self.constrain_latent_embedding and self.constrain_latent_key in self.adata.obsm.keys():
            P = self.adata.obsm[self.constrain_latent_key]
            if additional_batch_categories is not None:
                if batch_categories is not None and label_categories is not None and additional_label_categories is not None:
                    _dataset = list(zip(X, P, batch_categories, label_categories, *additional_label_categories, *additional_batch_categories))
                elif batch_categories is not None and label_categories is not None:
                    _dataset = list(zip(X, P, batch_categories, label_categories, *additional_batch_categories))
                elif batch_categories is not None:
                    _dataset = list(zip(X, P, batch_categories, *additional_batch_categories))
                elif label_categories is not None:
                    _dataset = list(zip(X, P, label_categories, *additional_batch_categories))
                else:
                    _dataset = list(zip(X, P, *additional_batch_categories))
            else:
                if batch_categories is not None and label_categories is not None and additional_label_categories is not None:
                    _dataset = list(zip(X, P, batch_categories, label_categories, *additional_label_categories))
                elif batch_categories is not None and label_categories is not None:
                    _dataset = list(zip(X, P, batch_categories, label_categories))
                elif batch_categories is not None:
                    _dataset = list(zip(X, P, batch_categories))
                elif label_categories is not None:
                    _dataset = list(zip(X, P, label_categories))
                else:
                    _dataset = list(zip(X, P))
        else:
            if additional_batch_categories is not None:
                if batch_categories is not None and label_categories is not None and additional_label_categories is not None:
                    _dataset = list(zip(X, batch_categories, label_categories, *additional_label_categories, *additional_batch_categories))
                elif batch_categories is not None and label_categories is not None:
                    _dataset = list(zip(X, batch_categories, label_categories, *additional_batch_categories))
                elif batch_categories is not None:
                    _dataset = list(zip(X, batch_categories, *additional_batch_categories))
                elif label_categories is not None:
                    _dataset = list(zip(X, label_categories, *additional_batch_categories))
                else:
                    _dataset = list(zip(X, *additional_batch_categories))
            else:
                if batch_categories is not None and label_categories is not None and additional_label_categories is not None:
                    _dataset = list(zip(X, batch_categories, label_categories, *additional_label_categories))
                elif batch_categories is not None and label_categories is not None:
                    _dataset = list(zip(X, batch_categories, label_categories))
                elif batch_categories is not None:
                    _dataset = list(zip(X, batch_categories))
                elif label_categories is not None:
                    _dataset = list(zip(X, label_categories))
                else:
                    _dataset = list(X)


        _shuffle_indices = list(range(len(_dataset)))
        np.random.shuffle(_shuffle_indices)
        self._dataset = np.array([_dataset[i] for i in _shuffle_indices])

        self._shuffle_indices = np.array(
            [x for x, _ in sorted(zip(range(len(_dataset)), _shuffle_indices), key=lambda x: x[1])]
        )

        self._shuffled_indices_inverse = _shuffle_indices


    def as_dataloader(
        self,
        subset_indices: Union[torch.tensor, np.ndarray] = None,
        n_per_batch: int = 128,
        train_test_split: bool = False,
        random_seed: bool = 42,
        validation_split: bool = .2,
        shuffle: bool = True,
    ):
        indices = subset_indices if subset_indices is not None else self._indices
        np.random.seed(random_seed)
        if shuffle:
            np.random.shuffle(indices)
        if train_test_split:
            split = int(np.floor(validation_split * self._n_record))
            if split % n_per_batch == 1:
                n_per_batch -= 1
            elif (self._n_record - split) % n_per_batch == 1:
                n_per_batch += 1
            train_indices, val_indices = indices[split:], indices[:split]
            train_sampler = SubsetRandomSampler(train_indices)
            valid_sampler = SubsetRandomSampler(val_indices)
            return DataLoader(indices, n_per_batch,  sampler = train_sampler), DataLoader(indices, n_per_batch, sampler = valid_sampler)
        if len(indices) % n_per_batch == 1:
            n_per_batch -= 1
        return DataLoader(indices, n_per_batch, shuffle = shuffle)

    def _normalize_data(self, X, after=None, copy=True):
        X = X.clone() if copy else X
        X = X.to(torch.float32)  # Check if torch.float64 should be used
        counts = X.sum(axis=1)
        counts_greater_than_zero = counts[counts > 0]
        after = torch.median(counts_greater_than_zero, dim=0).values if after is None else after
        counts += counts == 0
        counts = counts / after
        X /= counts.unsqueeze(1)
        return X
    

    def encode(self, X: torch.Tensor, batch_index: torch.Tensor = None, eps: float = 1e-4):
        # Encode for hidden space
        # if batch_index is not None and self.inject_batch:
        #    X = torch.hstack([X, batch_index])
        libsize = torch.log(X.sum(1))
        if self.reconstruction_method == 'zinb':
            if self.total_variational:
                X = self._normalize_data(X, after=1e4, copy=True)
            if self.log_variational:
                X = torch.log(1+X)
        q = self.encoder.encode(torch.hstack([X,libsize.unsqueeze(1)])) if self.encode_libsize else self.encoder.encode(X)
        q_mu = self.z_mean_fc(q)
        q_var = torch.exp(self.z_var_fc(q)) + eps
        z = Normal(q_mu, q_var.sqrt()).rsample()
        H = dict(
            q = q,
            q_mu = q_mu,
            q_var = q_var,
            z = z
        )

        return H

    def decode(self,
        H: Mapping[str, torch.tensor],
        lib_size:torch.tensor,
        batch_index: torch.Tensor = None,
        label_index: torch.Tensor = None,
        additional_batch_index: torch.Tensor = None,
        eps: float = 1e-4
    ):
        z = H["z"] # cell latent representation

        if additional_batch_index is not None and self.inject_additional_batch:
            if batch_index is not None and label_index is not None and self.inject_batch and self.inject_label:
                z = torch.hstack([z, batch_index, label_index, *additional_batch_index])
            elif batch_index is not None and self.inject_batch:
                z = torch.hstack([z, batch_index, *additional_batch_index])
            elif label_index is not None and self.inject_label:
                z = torch.hstack([z, label_index, *additional_batch_index])
        else:
            if batch_index is not None and label_index is not None and self.inject_batch and self.inject_label:
                z = torch.hstack([z, batch_index, label_index])
            elif batch_index is not None and self.inject_batch:
                z = torch.hstack([z, batch_index])
            elif label_index is not None and self.inject_label:
                z = torch.hstack([z, label_index])

        # eps to prevent numerical overflow and NaN gradient
        px = self.decoder(z)
        h = None

        px_rna_scale = self.px_rna_scale_decoder(px)
        if self.decode_libsize and not self.reconstruction_method == 'mse':
            px_rna_scale_final = px_rna_scale * lib_size.unsqueeze(1)
        elif self.reconstruction_method == 'mse':
            px_rna_scale_final = torch.log(px_rna_scale * 1e4 + 1)
        else:
            px_rna_scale_final = px_rna_scale

        if self.dispersion == "gene-cell":
            px_rna_rate = self.px_rna_rate_decoder(px) ## In logits
        elif self.dispersion == "gene-batch":
            px_rna_rate = F.linear(one_hot(batch_index, self.n_batch), self.px_rate)
        elif self.dispersion == "gene":
            px_rna_rate = self.px_rate

        px_rna_dropout = self.px_rna_dropout_decoder(px)  ## In logits

        R = dict(
            h = h,
            px = px,
            px_rna_scale_orig = px_rna_scale,
            px_rna_scale = px_rna_scale_final,
            px_rna_rate = px_rna_rate,
            px_rna_dropout = px_rna_dropout
        )
        return R

    def forward(
        self,
        X: torch.Tensor,
        lib_size: torch.Tensor,
        batch_index: torch.Tensor = None,
        label_index: torch.Tensor = None,
        additional_label_index: torch.Tensor = None,
        additional_batch_index: torch.Tensor = None,
        P: torch.Tensor = None,
        reduction: str = "sum",
        compute_mmd: bool = False
    ):
        H = self.encode(X, batch_index)
        q_mu = H["q_mu"]
        q_var = H["q_var"]
        mean = torch.zeros_like(q_mu)
        scale = torch.ones_like(q_var)
        kldiv_loss = kld(Normal(q_mu, q_var.sqrt()), Normal(mean, scale)).sum(dim = 1)
        prediction_loss = torch.tensor(0., device=self.device)
        additional_prediction_loss = torch.tensor(0., device=self.device)
        R = self.decode(H, lib_size, batch_index, label_index, additional_batch_index)

        if self.reconstruction_method == 'zinb':
            reconstruction_loss = LossFunction.zinb_reconstruction_loss(
                X,
                mu = R['px_rna_scale'],
                theta = R['px_rna_rate'].exp(),
                gate_logits = R['px_rna_dropout'],
                reduction = reduction
            )
        elif self.reconstruction_method == 'zg':
            reconstruction_loss = LossFunction.zi_gaussian_reconstruction_loss(
                X,
                mean=R['px_rna_scale'],
                variance=R['px_rna_rate'].exp(),
                gate_logits=R['px_rna_dropout'],
                reduction=reduction
            )
        elif self.reconstruction_method == 'mse':
            X_norm = self._normalize_data(X, after=1e4)
            X_norm = torch.log(X_norm + 1)
            reconstruction_loss = nn.functional.mse_loss(
                X_norm,
                R['px_rna_scale'],
                reduction=reduction
            )

        if self.n_label > 0:
            criterion = nn.CrossEntropyLoss(weight=self.label_category_weight)

            prediction = self.fc(H['z'])

            if self.new_adata_code and self.new_adata_code in label_index:
                prediction_index = (label_index != self.new_adata_code).squeeze()
                prediction_loss = criterion(prediction[prediction_index], one_hot(label_index[prediction_index], self.n_label))
            else:
                prediction_loss = criterion(prediction, one_hot(label_index, self.n_label))

        if self.n_additional_label is not None:
            prediction_loss = prediction_loss * self.additional_label_weight[0]
            for e,i in enumerate(self.n_additional_label):
                criterion = nn.CrossEntropyLoss(weight=self.additional_label_category_weight[e])
                additional_prediction = self.additional_fc[e](H['z'])
                if self.additional_new_adata_code[e] and self.additional_new_adata_code[e] in additional_label_index[e]:
                    additional_prediction_index = (additional_label_index[e] != self.additional_new_adata_code[e]).squeeze()

                    additional_prediction_loss += criterion(
                        additional_prediction[additional_prediction_index],
                        one_hot(additional_label_index[e][additional_prediction_index], i) * self.additional_label_weight[e+1]
                    )
                else:
                    additional_prediction_loss += criterion(additional_prediction, one_hot(additional_label_index[e], i)) * self.additional_label_weight[e+1]

        latent_constrain = torch.tensor(0.)
        if self.constrain_latent_embedding and P is not None:
            # Constrains on cells with no PCA information will be ignored
            latent_constrain_mask = P.mean(1) != 0
            if self.constrain_latent_method == 'mse':
                latent_constrain = (
                    nn.MSELoss(reduction='none')(P, q_mu).sum(1) * latent_constrain_mask
                ).sum() / len(list(filter(lambda x: x != 0, P.detach().cpu().numpy().mean(1))))
            elif self.constrain_latent_method == 'normal':
                latent_constrain = (
                    kld(Normal(q_mu, q_var.sqrt()), Normal(P, torch.ones_like(P))).sum(1) * latent_constrain_mask
                ).sum() / len(list(filter(lambda x: x != 0, P.detach().cpu().numpy().mean(1))))

        mmd_loss = torch.tensor(0.)
        if self.mmd_key is not None and compute_mmd:
            if self.mmd_key == 'batch':
                mmd_loss = self.mmd_loss(
                    H['q_mu'],
                    batch_index.detach().cpu().numpy(),
                    dim=1
                )
            elif self.mmd_key == 'additional_batch':
                for i in range(len(self.additional_batch_keys)):
                    mmd_loss_st += self.mmd_loss(
                        H['q_mu'], 
                        additional_batch_index[i].detach().cpu().numpy(),
                        dim=1
                    )
            elif self.mmd_key == 'both':
                mmd_loss_st = self.mmd_loss(
                    H['q_mu'], 
                    batch_index.detach().cpu().numpy(),
                    dim=1
                )
                for i in range(len(self.additional_batch_keys)):
                    mmd_loss_st += self.mmd_loss(
                        H['q_mu'], 
                        additional_batch_index[i].detach().cpu().numpy(),
                        dim=1
                    )

        loss_record = {
            "reconstruction_loss": reconstruction_loss,
            "prediction_loss": prediction_loss,
            "additional_prediction_loss": additional_prediction_loss,
            "kldiv_loss": kldiv_loss,
            "mmd_loss": mmd_loss,
            "latent_constrain_loss": latent_constrain
        }
        return H, R, loss_record

    def calculate_metric(self, X_test, kl_weight, pred_weight, mmd_weight, reconstruction_reduction):
        epoch_total_loss = 0
        epoch_reconstruction_loss = 0
        epoch_kldiv_loss = 0
        epoch_prediction_loss = 0
        epoch_mmd_loss = 0
        b = 0
        with torch.no_grad():
            for b, X in enumerate(X_test):
                X, P, batch_index, label_index, additional_label_index, additional_batch_index, lib_size = self._prepare_batch(X)

                H, R, L = self.forward(
                    X,
                    lib_size,
                    batch_index,
                    label_index,
                    additional_label_index,
                    additional_batch_index,
                    P,
                    reduction=reconstruction_reduction,
                    compute_mmd = mmd_weight > 0
                )
                reconstruction_loss = L['reconstruction_loss']
                prediction_loss = pred_weight * L['prediction_loss']
                additional_prediction_loss = pred_weight * L['additional_prediction_loss']
                kldiv_loss = kl_weight * L['kldiv_loss']
                mmd_loss = mmd_weight * L['mmd_loss']

                avg_reconstruction_loss = reconstruction_loss.sum()
                avg_kldiv_loss = kldiv_loss.sum()
                avg_mmd_loss = mmd_loss

                epoch_reconstruction_loss += avg_reconstruction_loss.item()
                epoch_kldiv_loss += avg_kldiv_loss.item()
                if self.n_label > 0:
                    epoch_prediction_loss += prediction_loss.sum().item()
                if self.n_additional_label is not None:
                    epoch_prediction_loss += additional_prediction_loss.sum().item()

                epoch_mmd_loss += avg_mmd_loss
                epoch_total_loss += (avg_reconstruction_loss + avg_kldiv_loss + avg_mmd_loss).item()
        return {
            "epoch_reconstruction_loss":  epoch_reconstruction_loss / (b+1),
            "epoch_kldiv_loss": epoch_kldiv_loss / (b+1),
            "epoch_mmd_loss": epoch_mmd_loss / (b+1),
            "epoch_prediction_loss": epoch_prediction_loss / (b+1),
            "epoch_total_loss": epoch_total_loss / (b+1),
        }


    def fit(self,
        max_epoch: Optional[int] = None,
        n_per_batch:int = 128,
        kl_weight: float = 1.,
        pred_weight: float = 1.,
        mmd_weight: float = 1.,
        constrain_weight: float = 1.,
        optimizer_parameters: Iterable = None,
        validation_split: float = .2,
        lr: bool = 5e-5,
        lr_schedule: bool = False,
        lr_factor: float = 0.6,
        lr_patience: int = 30,
        lr_threshold: float = 0.0,
        lr_min: float = 1e-6,
        n_epochs_kl_warmup: Union[int, None] = 400,
        weight_decay: float = 1e-6,
        random_seed: int = 12,
        subset_indices: Union[torch.tensor, np.ndarray] = None,
        pred_last_n_epoch: int = 10,
        pred_last_n_epoch_fconly: bool = False,
        reconstruction_reduction: str = 'sum',
    ):
        """
        Fit the model.
        
        :param max_epoch: int. Maximum number of epoch to train the model. If not provided, the model will be trained for 400 epochs or 20000 / n_record * 400 epochs as default.
        :param n_per_batch: int. Number of cells per batch.
        :param kl_weight: float. (Maximum) weight of the KL divergence loss.
        :param pred_weight: float. weight of the prediction loss.
        :param mmd_weight: float. weight of the mmd loss. ignored if mmd_key is None
        :param constrain_weight: float. weight of the constrain loss. ignored if constrain_latent_embedding is False. 
        :param optimizer_parameters: Iterable. Parameters to be optimized. If not provided, all parameters will be optimized.
        :param validation_split: float. Percentage of data to be used as validation set.
        :param lr: float. Learning rate.
        :param lr_schedule: bool. Whether to use learning rate scheduler.
        :param lr_factor: float. Factor to reduce learning rate.
        :param lr_patience: int. Number of epoch to wait before reducing learning rate.
        :param lr_threshold: float. Threshold to trigger learning rate reduction.
        :param lr_min: float. Minimum learning rate.
        :param n_epochs_kl_warmup: int. Number of epoch to warmup the KL divergence loss (deterministic warm-up
        of the KL-term).
        :param weight_decay: float. Weight decay (L2 penalty).
        :param random_seed: int. Random seed.
        :param subset_indices: Union[torch.tensor, np.ndarray]. Indices of cells to be used for training. If not provided, all cells will be used.
        :param pred_last_n_epoch: int. Number of epoch to train the prediction layer only.
        :param pred_last_n_epoch_fconly: bool. Whether to train the prediction layer only.
        :param reconstruction_reduction: str. Reduction method for reconstruction loss. Can be 'sum' or 'mean'.
        """
        self.train()
        if max_epoch is None:
            max_epoch = np.min([round((20000 / self._n_record ) * 400), 400])
            mt(f"max_epoch is not provided, setting max_epoch to {max_epoch}")
        if n_epochs_kl_warmup:
            n_epochs_kl_warmup = min(max_epoch, n_epochs_kl_warmup)
            kl_warmup_gradient = kl_weight / n_epochs_kl_warmup
            kl_weight_max = kl_weight
            kl_weight = 0.

        if optimizer_parameters is None:
            optimizer = optim.AdamW(self.parameters(), lr, weight_decay=weight_decay)
        else:
            optimizer = optim.AdamW(optimizer_parameters, lr, weight_decay=weight_decay)

        scheduler = ReduceLROnPlateau(
            optimizer,
            patience=lr_patience,
            factor=lr_factor,
            threshold=lr_threshold,
            min_lr=lr_min,
            threshold_mode="abs",
            verbose=True,
        ) if lr_schedule else None

        labels=None

        best_state_dict = None
        best_score = 0
        current_score = 0
        pbar = get_tqdm()(range(max_epoch), desc="Epoch", bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')
        loss_record = {
            "epoch_reconstruction_loss": 0,
            "epoch_kldiv_loss": 0,
            "epoch_prediction_loss": 0,
            "epoch_mmd_loss": 0,
            "epoch_total_loss": 0
        }

        epoch_total_loss_list = []
        epoch_reconstruction_loss_list = []
        epoch_kldiv_loss_list = []
        epoch_prediction_loss_list = []
        epoch_mmd_loss_list = []


        for epoch in range(1, max_epoch+1):
            self._trained = True
            pbar.desc = "Epoch {}".format(epoch)
            epoch_total_loss = 0
            epoch_reconstruction_loss = 0
            epoch_kldiv_loss = 0
            epoch_prediction_loss = 0
            epoch_mmd_loss = 0
            X_train, X_test = self.as_dataloader(
                n_per_batch=n_per_batch,
                train_test_split = True,
                validation_split = validation_split,
                random_seed=random_seed,
                subset_indices=subset_indices
            )
            if self.n_label > 0 and epoch+1 == max_epoch - pred_last_n_epoch:
                mt("saving transcriptome only state dict")
                self.gene_only_state_dict = deepcopy(self.state_dict())
                if  pred_last_n_epoch_fconly:
                    optimizer = optim.AdamW(chain(self.att.parameters(), self.fc.parameters()), lr, weight_decay=weight_decay)

            for b, X in enumerate(X_train):
                X, P, batch_index, label_index, additional_label_index, additional_batch_index, lib_size = self._prepare_batch(X)

                H, R, L = self.forward(
                    X,
                    lib_size,
                    batch_index,
                    label_index,
                    additional_label_index,
                    additional_batch_index,
                    P,
                    reduction=reconstruction_reduction,
                    compute_mmd = mmd_weight > 0
                )

                reconstruction_loss = L['reconstruction_loss']
                prediction_loss = pred_weight * L['prediction_loss']
                additional_prediction_loss = pred_weight * L['additional_prediction_loss']
                kldiv_loss = kl_weight * L['kldiv_loss']
                mmd_loss = mmd_weight * L['mmd_loss']

                avg_reconstruction_loss = reconstruction_loss.sum()  / n_per_batch
                avg_kldiv_loss = kldiv_loss.sum()  / n_per_batch
                avg_mmd_loss = mmd_loss / n_per_batch

                epoch_reconstruction_loss += avg_reconstruction_loss.item()
                epoch_kldiv_loss += avg_kldiv_loss.item()
                if self.n_label > 0:
                    epoch_prediction_loss += prediction_loss.sum().item()

                if epoch > max_epoch - pred_last_n_epoch:
                    loss = avg_reconstruction_loss + avg_kldiv_loss + avg_mmd_loss + prediction_loss.sum() + additional_prediction_loss.sum()
                else:
                    loss = avg_reconstruction_loss + avg_kldiv_loss + avg_mmd_loss

                if self.constrain_latent_embedding:
                    loss += constrain_weight * L['latent_constrain_loss']

                epoch_total_loss += loss.item()
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                pbar.set_postfix({
                    'reconst': '{:.2e}'.format(loss_record["epoch_reconstruction_loss"]),
                    'kldiv': '{:.2e}'.format(loss_record["epoch_kldiv_loss"]),
                    'pred': '{:.2e}'.format(loss_record["epoch_prediction_loss"]),
                    'mmd': '{:.2e}'.format(loss_record["epoch_mmd_loss"]),
                    'step': f'{b} / {len(X_train)}'
                })
            loss_record = self.calculate_metric(X_test, kl_weight, pred_weight, mmd_weight, reconstruction_reduction)
            if lr_schedule:
                scheduler.step(loss_record["epoch_total_loss"])
            pbar.set_postfix({
                'reconst': '{:.2e}'.format(loss_record["epoch_reconstruction_loss"]),
                'kldiv': '{:.2e}'.format(loss_record["epoch_kldiv_loss"]),
                'pred': '{:.2e}'.format(loss_record["epoch_prediction_loss"]),
                'mmd': '{:.2e}'.format(loss_record["epoch_mmd_loss"]),
            })
            epoch_total_loss_list.append(epoch_total_loss)
            epoch_reconstruction_loss_list.append(epoch_reconstruction_loss)
            epoch_kldiv_loss_list.append(epoch_kldiv_loss)
            epoch_prediction_loss_list.append(epoch_prediction_loss)
            epoch_mmd_loss_list.append(epoch_mmd_loss)
            pbar.update(1)
            if n_epochs_kl_warmup:
                kl_weight = min( kl_weight + kl_warmup_gradient, kl_weight_max)
            random_seed += 1
        if current_score < best_score:
            mt("restoring state dict with best performance")
            self.load_state_dict(best_state_dict)
        pbar.close()
        self.trained_state_dict = deepcopy(self.state_dict())
        return dict(
            epoch_total_loss_list=epoch_total_loss_list,   
            epoch_reconstruction_loss_list=epoch_reconstruction_loss_list,
            epoch_kldiv_loss_list=epoch_kldiv_loss_list,
            epoch_prediction_loss_list=epoch_prediction_loss_list,
            epoch_mmd_loss_list=epoch_mmd_loss_list
        )
    

    @torch.no_grad()
    def predict_labels(
        self,
        n_per_batch: int = 128,
        return_pandas: bool = False,
        show_progress: bool = True
    ) -> List:
        """
        Predict labels from trained model. 

        :param n_per_batch: int. Number of cells for each mini-batch during inference.
        :param return_pandas: bool. return a pandas DataFrame if True else return a pytorch tensor.
        :param show_progress: bool. Show progress bar of total progress.
        """
        self.eval()
        X = self.as_dataloader(subset_indices = list(range(len(self._dataset))), shuffle=False, n_per_batch=n_per_batch)
        predictions = []
        additional_predictions = []
        if show_progress:
            pbar = get_tqdm()(X, desc="Predicting Labels", bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')

        for x in X:
            x = self._dataset[x.cpu().numpy()]
            batch_index = None
            label_index = None
            if self.n_batch > 0 or self.n_label > 0:
                if not isinstance(x, Iterable) and len(x) > 1:
                    raise ValueError()
                if self.n_batch > 0 and self.n_label > 0:
                    X, batch_index, label_index = get_k_elements(x,0), get_k_elements(x,1), get_k_elements(X,2)
                elif self.n_batch > 0:
                    X, batch_index = get_k_elements(x,0), get_k_elements(x,1)
                elif self.n_label > 0:
                    X, label_index = get_k_elements(x,0), get_k_elements(x,1)

            if self.n_label > 0:
                label_index = torch.tensor(label_index)
                if not isinstance(label_index, torch.FloatTensor):
                    label_index = label_index.type(torch.FloatTensor)
                label_index = label_index.to(self.device).unsqueeze(1)
            if self.n_batch > 0:
                batch_index = torch.tensor(batch_index)
                if not isinstance(batch_index, torch.FloatTensor):
                    batch_index = batch_index.type(torch.FloatTensor)
                batch_index = batch_index.to(self.device).unsqueeze(1)

            X = torch.tensor(np.vstack(list(map(lambda x: x.toarray() if issparse(x) else x, X))))
            if not isinstance(X, torch.FloatTensor):
                X = X.type(torch.FloatTensor)
            X = X.to(self.device)

            H = self.encode(X, batch_index if batch_index != None else None)
            prediction = self.fc(H['z'])
            predictions.append(prediction.detach().cpu())

            if self.n_additional_label is not None:
                additional_prediction = [None] * len(self.n_additional_label)
                for i in range(len(self.n_additional_label)):
                    additional_prediction[i] = self.additional_fc[i](H['z']).detach().cpu()
                additional_predictions.append(additional_prediction)

            if show_progress:
                pbar.update(1)
        if show_progress:
            pbar.close()

        predictions = torch.vstack(predictions)[self._shuffle_indices]
        predictions_argmax = torch.argmax(predictions, dim=1)
        predictions_argmax = list(map(lambda x:
            self.label_category.categories[x],
            predictions_argmax.detach().cpu().numpy()
        ))

        if return_pandas:
            predictions_argmax = pd.DataFrame(predictions_argmax)
            predictions_argmax.columns = [self.label_key]
            return predictions_argmax

        if self.n_additional_label is not None:
            additional_predictions_result = [None] * len(self.n_additional_label)
            additional_predictions_result_argmax = [None] * len(self.n_additional_label)
            for i in range(len(self.n_additional_label)):
                additional_predictions_ = torch.vstack([additional_predictions[x][i] for x in range(len(additional_predictions))]) [self._shuffle_indices]
                additional_predictions_result_argmax[i] = np.argmax(additional_predictions_, axis=1)
                additional_predictions_result_argmax[i] = list(map(lambda x:
                    self.additional_label_category[i].categories[x],
                    additional_predictions_result_argmax[i].numpy()
                ))
            if return_pandas:
                additional_predictions_result_argmax = pd.DataFrame(additional_predictions_result_argmax).T
                additional_predictions_result_argmax.columns = self.additional_label_keys
                return pd.concat([predictions_argmax, additional_predictions_result_argmax], axis=1)

            return predictions, additional_predictions

        return predictions


    @torch.no_grad()
    def get_latent_embedding(
        self,
        latent_key: Literal["z", "q_mu"] = "q_mu",
        n_per_batch: int = 128,
        show_progress: bool = True
    ) -> np.ndarray:
        self.eval()
        X = self.as_dataloader(
            subset_indices = list(range(len(self._dataset))), 
            shuffle=False, 
            n_per_batch=n_per_batch
        )
        if isinstance(latent_key, str):
            Zs = []
        elif isinstance(latent_key, Iterable):
            Zs = [[] for _ in range(len(latent_key))]
        if show_progress:
            pbar = get_tqdm()(X, desc="Latent Embedding", bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')

        for x in X:
            X, P, batch_index, label_index, additional_label_index, additional_batch_index, lib_size = self._prepare_batch(x)

            H = self.encode(X, batch_index if batch_index != None else None)
            if isinstance(latent_key, str):
                Zs.append(H[latent_key].detach().cpu().numpy())
            elif isinstance(latent_key, Iterable):
                for i in range(len(latent_key)):
                    Zs[i].append(H[latent_key[i]].detach().cpu().numpy())
            if show_progress:
                pbar.update(1)
        if show_progress:
            pbar.close()
        if isinstance(latent_key, str):
            return np.vstack(Zs)[self._shuffle_indices]
        elif isinstance(latent_key, Iterable):
            return [np.vstack(Z)[self._shuffle_indices] for Z in Zs]

    @torch.no_grad()
    def get_reconstructed_expression(self, k = 'px_rna_scale_orig', n_per_batch=256,show_progress=True) -> np.ndarray:
        self.eval()
        Zs = []
        X = self.as_dataloader(subset_indices = list(range(len(self._dataset))), shuffle=False, n_per_batch=n_per_batch)
        predictions = []
        additional_predictions = []
        if show_progress:
            pbar = get_tqdm()(X, desc="Predicting Labels", bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')

        for x in X:
            X, P, batch_index, label_index, additional_label_index, additional_batch_index, lib_size = self._prepare_batch(x)

            _,R,_ = self.forward(
                X,
                lib_size,
                batch_index,
                label_index,
                additional_label_index,
                additional_batch_index,
                P,
            )
            Zs.append(R[k].detach().cpu().numpy())
            if show_progress:
                pbar.update(1)
        return np.vstack(Zs)[self._shuffle_indices]


    def to(self, device:str):
        super(scAtlasVAE, self).to(device)
        self.device=device
        return self

    def transfer(self,
        new_adata: sc.AnnData,
        batch_key: str,
        concat_with_original: bool = True,
        fraction_of_original: Optional[float] = None,
        times_of_new: Optional[float] = None
    ):
        new_batch_category = new_adata.obs[batch_key]
        original_batch_dim = self.batch_hidden_dim
        new_n_batch = len(np.unique(new_batch_category))

        if self.batch_embedding == "embedding":
            original_embedding_weight = self.decoder.cat_embedding[0].weight

        new_adata.obs[self.batch_key] = new_adata.obs[batch_key]

        original_batch_categories = self.batch_category.categories

        if fraction_of_original is not None:
            old_adata = random_subset_by_key_fast(
                self.adata,
                key = batch_key,
                n = int(len(self.adata) * fraction_of_original)
            )
        elif times_of_new is not None:
            old_adata = random_subset_by_key_fast(
                self.adata,
                key = batch_key,
                n = int(len(new_adata) * times_of_new)
            )
        else:
            old_adata = self.adata

        old_adata.obs['_transfer_label'] = 'reference'
        new_adata.obs['_transfer_label'] = 'query'

        if concat_with_original:
            self.adata = sc.concat([old_adata, new_adata])
        else:
            self.adata = new_adata

        self.initialize_dataset()

        if self.batch_embedding == "onehot":
            self.batch_hidden_dim = self.n_batch

        if self.n_additional_batch_ is not None and self.inject_additional_batch:
            if self.n_batch > 0 and self.n_label > 0 and self.inject_batch and self.inject_label:
                decoder_n_cat_list = [self.n_batch, self.n_label, *self.n_additional_batch]
            elif self.n_batch > 0 and self.inject_batch:
                decoder_n_cat_list = [self.n_batch, *self.n_additional_batch]
            elif self.n_label > 0 and self.inject_label:
                decoder_n_cat_list = [self.n_label, *self.n_additional_batch]
            else:
                decoder_n_cat_list = None
        else:
            if self.n_batch > 0 and self.n_label > 0 and self.inject_batch and self.inject_label:
                decoder_n_cat_list = [self.n_batch, self.n_label]
            elif self.n_batch > 0 and self.inject_batch:
                decoder_n_cat_list = [self.n_batch]
            elif self.n_label > 0 and self.inject_label:
                decoder_n_cat_list = [self.n_label]
            else:
                decoder_n_cat_list = None


        self.decoder_n_cat_list = decoder_n_cat_list

        original_weight = torch.tensor(self.decoder._fclayer[0].weight)

        self.decoder = FCLayer(
            in_dim = self.n_latent,
            out_dim = self.n_hidden,
            n_cat_list = self.decoder_n_cat_list,
            cat_dim = self.batch_hidden_dim,
            cat_embedding = self.batch_embedding,
            use_layer_norm=False,
            use_batch_norm=True,
            dropout_rate=0,
            device=self.device
        )


        if self.batch_embedding == 'embedding':
            new_embedding = nn.Embedding(self.n_batch + new_n_batch, self.batch_hidden_dim).to(self.device)
            original_category_index = [list(self.batch_category.categories).index(x) for x in original_batch_categories]
            new_embedding_weight = new_embedding.weight.detach()
            new_embedding_weight[original_category_index] = original_embedding_weight.detach()
            new_embedding.weight = nn.Parameter(new_embedding_weight)
            new_embedding = new_embedding.to(self.device)
            self.decoder.cat_embedding[0] = new_embedding

        new_weight = torch.tensor(self.decoder._fclayer[0].weight)
        new_weight[:,:(self.n_latent + original_batch_dim)] = original_weight[:,:(self.n_latent + original_batch_dim)]
        self.decoder._fclayer[0].weight = nn.Parameter(new_weight)
        self.to(self.device)


    def transfer_label(
        self,
        reference_adata: sc.AnnData,
        label_key: str,
        method: Literal['knn'] = 'knn',
        use_rep: str = 'X_gex',
        **method_kwargs
    ):
        """
        Transfer label from reference_adata to self.adata

        :param reference_adata: sc.AnnData
        :param label_key: str
        """
        s = set(reference_adata.obs.index)
        s = list(filter(lambda x: x in s, self.adata.obs.index))

        self.adata.obs[label_key] = np.nan
        ss = set(s)
        indices = list(map(lambda x: x in ss, self.adata.obs.index))
        self.adata.obs[label_key][indices] = reference_adata[s].obs[label_key]

        if 'X_gex' not in self.adata.obsm.keys():
            Z = self.get_latent_embedding()
            self.adata.obsm[use_rep] = Z

        if method == 'knn':
            from sklearn.neighbors import KNeighborsClassifier
            knn = KNeighborsClassifier(**method_kwargs)
            knn.fit(
                self.adata.obsm[use_rep][indices], self.adata.obs.loc[indices, label_key]
            )
            self.adata.obs.loc[self.adata.obs['_transfer_label'] == 'query', label_key] = knn.predict(
                self.adata.obsm[use_rep][self.adata.obs['_transfer_label'] == 'query']
            )
        else:
            raise NotImplementedError()

    def transfer_umap(
        self,
        reference_adata: sc.AnnData,
        label_key: str,
        method: Literal['retrain','knn'] = 'knn',
        use_rep: str = 'X_gex',
        **method_kwargs
    ):
        transfer_umap(
            reference_adata.obsm[use_rep],
            reference_adata.obsm['X_umap'],
            reference_adata.obsm[use_rep],
            method=method,

        )

    def _prepare_batch(self, X):
                P = None
                X = self._dataset[X.cpu().numpy()]
                batch_index, label_index, additional_label_index, additional_batch_index = None, None, None, None
                if self.n_batch > 0 or self.n_label > 0:
                    if not isinstance(X, Iterable) and len(X) > 1:
                        raise ValueError()
                    if self.n_additional_batch_ is not None:
                        if self.n_batch > 0 and self.n_label > 0 and self.n_additional_label is not None:
                            if self.constrain_latent_embedding:
                                X, P, batch_index, label_index, additional_label_index, additional_batch_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_k_elements(X,3),
                                    get_elements(X,4, len(self.n_additional_label)),
                                    get_last_k_elements(X,4+len(self.n_additional_label))
                                )
                            else:
                                X, batch_index, label_index, additional_label_index, additional_batch_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_elements(X,3, len(self.n_additional_label)),
                                    get_last_k_elements(X,3+len(self.n_additional_label))
                                )
                            additional_label_index = list(np.vstack(additional_label_index).T.astype(int))
                        elif self.n_batch > 0 and self.n_label > 0:
                            if self.constrain_latent_embedding:
                                X, P, batch_index, label_index, additional_batch_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_k_elements(X,3),
                                    get_last_k_elements(X,4)
                                )
                            else:
                                X, batch_index, label_index, additional_batch_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_last_k_elements(X,3)
                                )
                        elif self.n_batch > 0:
                            if self.constrain_latent_embedding:
                                X, P, batch_index, additional_batch_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_last_k_elements(X,3)
                                )
                            else:
                                X, batch_index, additional_batch_index = get_k_elements(X,0), get_k_elements(X,1),  get_last_k_elements(X,2)
                        elif self.n_label > 0:
                            if self.constrain_latent_embedding:
                                X, P, label_index, additional_batch_index = get_k_elements(X,0), get_k_elements(X,1), get_k_elements(X,2), get_last_k_elements(X,3)
                            else:
                                X, label_index, additional_batch_index = get_k_elements(X,0), get_k_elements(X,1), get_last_k_elements(X,2)
                        additional_batch_index = list(np.vstack(additional_batch_index).T.astype(int))
                    else:
                        if self.n_batch > 0 and self.n_label > 0 and self.n_additional_label is not None:
                            if self.constrain_latent_embedding:
                                X, P, batch_index, label_index, additional_label_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_k_elements(X,3),
                                    get_last_k_elements(X,4)
                                )
                            else:
                                X, batch_index, label_index, additional_label_index = (
                                    get_k_elements(X,0),
                                    get_k_elements(X,1),
                                    get_k_elements(X,2),
                                    get_last_k_elements(X,3)
                                )
                            additional_label_index = list(np.vstack(additional_label_index).T.astype(int))
                        elif self.n_batch > 0 and self.n_label > 0:
                            if self.constrain_latent_embedding:
                                X, P, batch_index, label_index = get_k_elements(X,0), get_k_elements(X,1), get_k_elements(X,2), get_k_elements(X,3)
                            else:
                                X, batch_index, label_index = get_k_elements(X,0), get_k_elements(X,1), get_k_elements(X,2)
                        elif self.n_batch > 0:
                            if self.constrain_latent_embedding:
                                X, P, batch_index = get_k_elements(X,0), get_k_elements(X,1),  get_k_elements(X,2)
                            else:
                                X, batch_index = get_k_elements(X,0), get_k_elements(X,1)
                        elif self.n_label > 0:
                            if self.constrain_latent_embedding:
                                X, P, label_index = get_k_elements(X,0), get_k_elements(X,1), get_k_elements(X,2)
                            else:
                                X, label_index = get_k_elements(X,0), get_k_elements(X,1)


                X = torch.tensor(np.vstack(list(map(lambda x: x.toarray() if issparse(x) else x, X))))
                if self.constrain_latent_embedding:
                    P = torch.tensor(np.vstack(P)).type(torch.FloatTensor).to(self.device)

                if self.n_label > 0:
                    label_index = torch.tensor(label_index)
                    if not isinstance(label_index, torch.FloatTensor):
                        label_index = label_index.type(torch.FloatTensor)
                    label_index = label_index.to(self.device).unsqueeze(1)

                if self.n_batch > 0:
                    batch_index = torch.tensor(batch_index)
                    if not isinstance(batch_index, torch.FloatTensor):
                        batch_index = batch_index.type(torch.FloatTensor)
                    batch_index = batch_index.to(self.device).unsqueeze(1)

                if self.n_additional_label is not None:
                    for i in range(len(additional_label_index)):
                        additional_label_index[i] = torch.tensor(additional_label_index[i])
                        if not isinstance(additional_label_index[i], torch.FloatTensor):
                            additional_label_index[i] = additional_label_index[i].type(torch.FloatTensor)
                        additional_label_index[i] = additional_label_index[i].to(self.device).unsqueeze(1)

                if self.n_additional_batch_ is not None:
                    for i in range(len(additional_batch_index)):
                        additional_batch_index[i] = torch.tensor(additional_batch_index[i])
                        if not isinstance(additional_batch_index[i], torch.FloatTensor):
                            additional_batch_index[i] = additional_batch_index[i].type(torch.FloatTensor)
                        additional_batch_index[i] = additional_batch_index[i].to(self.device).unsqueeze(1)

                if not isinstance(X, torch.FloatTensor):
                    X = X.type(torch.FloatTensor)
                if P is not None and not isinstance(P, torch.FloatTensor):
                    P = P.type(torch.FloatTensor)
                    P = P.to(self.device)
                X = X.to(self.device)

                lib_size = X.sum(1).to(self.device)
                return X, P, batch_index, label_index, additional_label_index, additional_batch_index, lib_size


import torch
from torch import nn, optim
from torch.nn import functional as F
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau
from typing import Union, Mapping, Iterable, List, Optional, Callable
from torch.distributions import Normal
from torch.utils.data import DataLoader
from torch import nn, einsum
import warnings

from einops.layers.torch import Rearrange, Reduce
from einops import rearrange, repeat

import numpy as np

from functools import partial

from ..utils._loss import LossFunction
from ..utils._logger import  mt
from ..utils._definitions import FCDEF
from ..utils._tensor_utils import one_hot
from ..utils._compat import Literal

import math
from scipy.stats import truncnorm
import torch
import torch.nn as nn

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

#https://github.com/Bjarten/early-stopping-pytorch/blob/master/pytorchtools.py
class EarlyStopping:
    """
    Early stops the training if validation loss doesn't improve after a given patience.
    >>> early_stopping = EarlyStopping(patience=patience, verbose=True)
    >>> early_stopping(valid_loss, model)
    """
    def __init__(self, patience=7, verbose=False, delta=0, path='checkpoint.pt', trace_func=print):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement. 
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
            trace_func (function): trace print function.
                            Default: print            
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.path = path
        self.trace_func = trace_func
    def __call__(self, val_loss, model):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
        elif score < self.best_score + self.delta:
            self.counter += 1
            self.trace_func(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss, model):
        '''Saves model when validation loss decrease.'''
        if self.verbose:
            self.trace_func(f'Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ...')
        torch.save(model.state_dict(), self.path)
        self.val_loss_min = val_loss




####################################################
#                    Linear                        #
####################################################

def _prod(nums):
    out = 1
    for n in nums:
        out = out * n
    return out

def _calculate_fan(linear_weight_shape, fan="fan_in"):
    fan_out, fan_in = linear_weight_shape

    if fan == "fan_in":
        f = fan_in
    elif fan == "fan_out":
        f = fan_out
    elif fan == "fan_avg":
        f = (fan_in + fan_out) / 2
    else:
        raise ValueError("Invalid fan option")

    return f

def _trunc_normal_init_(weights, scale=1.0, fan="fan_in"):
    shape = weights.shape
    f = _calculate_fan(shape, fan)
    scale = scale / max(1, f)
    a = -2
    b = 2
    std = math.sqrt(scale) / truncnorm.std(a=a, b=b, loc=0, scale=1)
    size = _prod(shape)
    samples = truncnorm.rvs(a=a, b=b, loc=0, scale=std, size=size)
    samples = np.reshape(samples, shape)
    with torch.no_grad():
        weights.copy_(torch.tensor(samples, device=weights.device))
        
def lecun_normal_init_(weights):
    _trunc_normal_init_(weights, scale=1.0)


def he_normal_init_(weights):
    _trunc_normal_init_(weights, scale=2.0)


def glorot_uniform_init_(weights):
    nn.init.xavier_uniform_(weights, gain=1)


def final_init_(weights):
    with torch.no_grad():
        weights.fill_(0.0)


def gating_init_(weights):
    with torch.no_grad():
        weights.fill_(0.0)


def normal_init_(weights):
    torch.nn.init.kaiming_normal_(weights, nonlinearity="linear")
    
class Linear(nn.Linear):
    """
    A Linear layer with built-in nonstandard initializations. Called just
    like torch.nn.Linear.
    """
    
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        bias: bool = True,
        init: Literal["default","final","gating","glorot","normal","relu"] = "default",
        init_fn: Optional[Callable[[torch.Tensor, torch.Tensor], None]] = None,
    ):
        """
        Args:
            in_dim:
                The final dimension of inputs to the layer
            out_dim:
                The final dimension of layer outputs
            bias:
                Whether to learn an additive bias. True by default
            init:
                The initializer to use. Choose from:

                "default": LeCun fan-in truncated normal initialization
                "relu": He initialization w/ truncated normal distribution
                "glorot": Fan-average Glorot uniform initialization
                "gating": Weights=0, Bias=1
                "normal": Normal initialization with std=1/sqrt(fan_in)
                "final": Weights=0, Bias=0

                Overridden by init_fn if the latter is not None.
            init_fn:
                A custom initializer taking weight and bias as inputs.
                Overrides init if not None.
        """
        super(Linear, self).__init__(in_dim, out_dim, bias=bias)

        if bias:
            with torch.no_grad():
                self.bias.fill_(0)

        with torch.no_grad():
            if init_fn is not None:
                init_fn(self.weight, self.bias)
            elif init == "default":
                lecun_normal_init_(self.weight)
            elif init == "final":
                final_init_(self.weight)
            elif init == "gating":
                gating_init_(self.weight)
                if bias:
                    self.bias.fill_(1.0)
            elif init == "glorot":
                glorot_uniform_init_(self.weight)
            elif init == "normal":
                normal_init_(self.weight)
            elif init == "relu":
                he_normal_init_(self.weight)
            else:
                raise ValueError("Invalid init string.")
            

class Sparsemax(nn.Module):
    """Sparsemax activation function.

    Pytorch implementation of Sparsemax function from:
    -- "From Softmax to Sparsemax: A Sparse Model of Attention and Multi-Label Classification"
    -- André F. T. Martins, Ramón Fernandez Astudillo (http://arxiv.org/abs/1602.02068)
    """
    def __init__(self, dim=None):
        """Initialize sparsemax activation
        
        Args:
            dim (int, optional): The dimension over which to apply the sparsemax function.
        """
        super(Sparsemax, self).__init__()

        self.dim = -1 if dim is None else dim

    def forward(self, input):
        """Forward function.

        Args:
            input (torch.Tensor): Input tensor. First dimension should be the batch size

        Returns:
            torch.Tensor: [batch_size x number_of_logits] Output tensor

        """
        # Sparsemax currently only handles 2-dim tensors,
        # so we reshape to a convenient shape and reshape back after sparsemax
        input = input.transpose(0, self.dim)
        original_size = input.size()
        input = input.reshape(input.size(0), -1)
        input = input.transpose(0, 1)
        dim = 1

        number_of_logits = input.size(dim)

        # Translate input by max for numerical stability
        input = input - torch.max(input, dim=dim, keepdim=True)[0].expand_as(input)

        # Sort input in descending order.
        # (NOTE: Can be replaced with linear time selection method described here:
        # http://stanford.edu/~jduchi/projects/DuchiShSiCh08.html)
        zs = torch.sort(input=input, dim=dim, descending=True)[0]
        range = torch.arange(start=1, end=number_of_logits + 1, step=1, device=input.device, dtype=input.dtype).view(1, -1)
        range = range.expand_as(zs)

        # Determine sparsity of projection
        bound = 1 + range * zs
        cumulative_sum_zs = torch.cumsum(zs, dim)
        is_gt = torch.gt(bound, cumulative_sum_zs).type(input.type())
        k = torch.max(is_gt * range, dim, keepdim=True)[0]

        # Compute threshold function
        zs_sparse = is_gt * zs

        # Compute taus
        taus = (torch.sum(zs_sparse, dim, keepdim=True) - 1) / k
        taus = taus.expand_as(input)

        # Sparsemax
        self.output = torch.max(torch.zeros_like(input), input - taus)

        # Reshape back to original shape
        output = self.output
        output = output.transpose(0, 1)
        output = output.reshape(original_size)
        output = output.transpose(0, self.dim)

        return output

    def backward(self, grad_output):
        """Backward function."""
        dim = 1

        nonzeros = torch.ne(self.output, 0)
        sum = torch.sum(grad_output * nonzeros, dim=dim) / torch.sum(nonzeros, dim=dim)
        self.grad_input = nonzeros * (grad_output - sum.expand_as(grad_output))

        return self.grad_input

        
class ReparameterizeLayerBase(nn.Module):
    """Base layer for VAE reparameterization"""
    def __init__(self) -> None:
        """
        @brief ReparameterizeLayer Defining methods for reparameterize tricks
        """
        super(ReparameterizeLayerBase, self).__init__()

    def reparameterize(
        self, 
        mu: torch.Tensor, 
        var: torch.Tensor
    ) -> torch.Tensor :
        untran_z = Normal(mu, var.sqrt()).rsample()
        return untran_z

    def reparameterize_transformation(
        self, 
        transfunc, 
        mu: torch.Tensor, 
        var: torch.Tensor
    ) -> torch.Tensor:
        z = self.reparameterize(mu, var)
        ztrans = transfunc(z)
        return ztrans, z

class MMDLayerBase:
    """Base layer for Maximum-mean descrepancy calculation"""
    def hierarchical_mmd_loss(
        self, 
        z: torch.Tensor, 
        cat: np.array, 
        hierarchical_weight: Iterable[float]
    ) -> torch.Tensor:
        if len(cat.shape) <= 1:
            raise ValueError("Illegal category array")
        if len(z) != cat.shape[0]:
            raise ValueError("Dimension of z {} should be equal to dimension of category {}".format(len(z), cat.shape[0]))
        if len(hierarchical_weight) != cat.shape[1]:
            raise ValueError("Dimension of hierarchical_weight {} should be equal to dimension of category {}".format(len(hierarchical_weight), cat.shape[1]))

        if cat.shape[1] < 2:
            cat = cat.flatten()
            return self.MMDLoss(z, cat)
        loss = 0
        zs = []
        for i in np.unique(cat[:, 0]):
            idx = list(map(lambda t:t[0], filter(lambda x:x[1] == i, enumerate(cat[:,0]))))
            loss += self.HierarchicalMMDLoss(z[idx], cat[idx,1:], hierarchical_weight[1:])
            zs.append(z[idx])
        for i in range(len(np.unique(cat[:,0]))):
            for j in range(i+1, len(np.unique(cat[:,0]))):
                loss += LossFunction.mmd_loss(
                    zs[i], zs[j]
                )
        return loss

    def mmd_loss(self, 
        z: torch.Tensor, 
        cat: np.array, 
        dim=-1,
        flavor: Literal['trvae','default'] = 'trvae'
    ) -> torch.Tensor:
        zs = []
        loss = 0
        for i in np.unique(cat):
            idx = list(map(lambda z:z[0], filter(lambda x:x[1] == i, enumerate(cat))))
            zs.append(z[idx])
        for i in range(len(np.unique(cat))):
            for j in range(i+1,len(np.unique(cat))):
                if flavor == 'trvae':
                    loss += LossFunction.mmd_loss_trvae(
                        zs[i], zs[j]
                    )
                else:
                    loss += LossFunction.mmd_loss(
                        zs[i], zs[j], dim=-1
                    )
        return loss

class FCLayer(nn.Module):
    """FCLayer Fully-Connected Layers for a neural network """
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        n_cat_list: Iterable[int] = None,
        cat_dim: int = 8,
        cat_embedding: Literal["embedding", "onehot"] = "onehot",
        bias: bool = True,
        dropout_rate: float = 0.1,
        use_batch_norm: bool = True,
        use_layer_norm: bool = False,
        activation_fn: nn.Module = nn.ReLU,
        activation_dim: int = None,
        device: str = "cuda"
    ):
        super(FCLayer, self).__init__()
        if n_cat_list is not None:
            # Categories
            if not all(map(lambda x: x > 1, n_cat_list)):
                warnings.warn("category list contains values less than 1")
            self.n_category = len(n_cat_list)
            self._cat_dim = cat_dim
            self._cat_embedding = cat_embedding
            if cat_embedding == "embedding":
                self.cat_dimension = self.n_category * cat_dim # Total dimension of categories using embedding
            else:
                self.cat_dimension = sum(n_cat_list) # Total dimension of categories using one-hot
            self.n_cat_list = n_cat_list
            if cat_embedding == "embedding":
                self.cat_embedding = nn.ModuleList(
                    [nn.Embedding(n, cat_dim) for n in n_cat_list]
                )
            else: 
                self.cat_embedding = [
                    partial(one_hot, n_cat=n) for n in n_cat_list
                ]

        else:
            # No categories will be included
            self.n_category = 0
            self.n_cat_list = None
        
        self._fclayer = nn.Sequential(
                *list(filter(lambda x:x, 
                        [
                            nn.Linear(in_dim, out_dim, bias=bias) 
                            if self.n_category == 0 
                            else nn.Linear(in_dim + self.cat_dimension, out_dim, bias=bias),
                            nn.BatchNorm1d(out_dim, momentum=0.01, eps=0.001) if use_batch_norm else None,
                            nn.LayerNorm(out_dim, elementwise_affine=False) if use_layer_norm else None,
                            activation_fn(dim=activation_dim) if activation_dim else activation_fn() if activation_fn else None,
                            nn.Dropout(p = dropout_rate) if dropout_rate > 0 else None
                        ]
                    )
                )
            )
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.device = device 

    def forward(self, X: torch.Tensor, cat_list: torch.Tensor =  None, return_category_embedding: bool = False):
        category_embedding = []
        if self.n_category > 0:
            if cat_list != None:
                if (len(cat_list) != self.n_category):
                    raise ValueError("Number of category list should be equal to {}".format(self.n_category))

                for i, (n_cat, cat) in enumerate(zip(self.n_cat_list, cat_list)):
                    assert(n_cat > 1)
                    if self._cat_embedding == "embedding":
                        category_embedding.append(self.cat_embedding[i](cat))
                    else: 
                        category_embedding.append(self.cat_embedding[i](cat.unsqueeze(0).T))
            else:
                if X.shape[1] != self.in_dim + self.n_category:
                    raise ValueError("Dimension of X should be equal to {} + {} but found {} if cat_list is provided".format(self.in_dim, self.n_category, X.shape[1]))
                cat_list = X[:, -self.n_category:].type(torch.LongTensor).T.to(self.device)
                for i, (n_cat, cat) in enumerate(zip(self.n_cat_list, cat_list)):
                    if self._cat_embedding == "embedding":
                        category_embedding.append(self.cat_embedding[i](cat))
                    else: 
                        category_embedding.append(self.cat_embedding[i](cat.unsqueeze(0).T))
               
            category_embedding = torch.hstack(category_embedding).to(self.device)
            if return_category_embedding:
                return self._fclayer(torch.hstack([X[:,:self.in_dim], category_embedding])), category_embedding
            else: 
                return self._fclayer(torch.hstack([X[:,:self.in_dim], category_embedding]))
        else:
            return self._fclayer(X)

    def to(self, device:str):
        super(FCLayer, self).to(device)
        self.device=device 
        return self

class PredictionLayerBase(nn.Module):
    """Prediction layer base """
    def __init__(self, *, in_dim:int, n_pred_category: int):
        super(PredictionLayerBase, self).__init__()
        self.in_dim = in_dim
        self.n_pred_category = n_pred_category
        self.decoder = FCLayer(
            in_dim = in_dim,
            out_dim = n_pred_category,
            bias = False,
            dropout_rate = 0,
            use_batch_norm = False,
            use_layer_norm = False,
            activation_fn = nn.ReLU,
        )

    def forward(self, X: torch.Tensor):
        return nn.Softmax(-1)( self.decoder(X) )

class SAE(nn.Module):
    ''' Stacked Autoencoders. 
        Fitting includes stacked fitting and fine-tuning:
            Fine-tuning step removes the decoder and use clustering method
            to fine-tune the encoder.
        parameters:
            dim:    int 
            stacks: Iterable[int]
            n_cat_list: Iterable[int]
            cat_dim: int
    '''
    def __init__(
            self, 
            dim:int, 
            stacks:Iterable[int] = [512, 128, 64], 
            n_cat_list: Iterable[int] = None,
            cat_dim: int = 8,
            bias: bool = True,
            dropout_rate: float = 0.1,
            use_batch_norm: bool = True,
            use_layer_norm: bool = False,
            cat_embedding: Literal["embedding", "onthot"] = "onehot",
            activation_fn: nn.Module = nn.ReLU,
            encode_only: bool = False,
            decode_only: bool = False,
            device="cuda"
    ):
        super(SAE, self).__init__()
        fcargs = dict(
            bias=bias, 
            dropout_rate=dropout_rate, 
            use_batch_norm=use_batch_norm, 
            use_layer_norm=use_layer_norm,
            activation_fn=activation_fn,
            device=device,
            cat_embedding = cat_embedding
        )
        self.dim = dim
        self.num_layers = len(stacks)
        self.n_cat_list = n_cat_list
        self.cat_dim = cat_dim
        self.n_category = len(n_cat_list) if n_cat_list != None else 0
        self.stacks = stacks
        self.out_dim = stacks[-1]
        layers = [None] * len(stacks)
        self.n_layers = len(stacks)
        if (encode_only & decode_only):
            raise ValueError("SAE instance cannot be both encode and decode only")
        for i,j in enumerate(stacks):
            if i == 0:
                layers[i] = [FCLayer(dim, 
                             stacks[i], 
                             n_cat_list, 
                             cat_dim,
                             **fcargs)
                             if not decode_only 
                             else None, 
                             FCLayer(stacks[i], dim, **fcargs) 
                             if not encode_only 
                             else None]
            else:
                layers[i] = [FCLayer(stacks[i-1], stacks[i], **fcargs)
                             if not decode_only 
                             else None, 
                             FCLayer(stacks[i], stacks[i-1], **fcargs) 
                             if not encode_only 
                             else None ]
        layers = [i for s in layers for i in s]
        self.layers = nn.ModuleList(layers)
        self.device = device
        self.loss = []
        self.encode_only = encode_only
        self.decode_only = decode_only

    def get_layer(self, codec:str, layer:int):
        i = 0 if codec == FCDEF.ENCODER else 1
        return self.layers[layer * 2 + i]

    def encode(self, x: torch.Tensor):
        '''
        encode features in the nth layer 
        '''
        if self.decode_only:
            raise TypeError("This is an decoder-only SAE instance")
        h = None
        for i in range(self.num_layers):
            layer = self.get_layer(FCDEF.ENCODER, i)
            if i == self.num_layers - 1:
                if i == 0:
                    h = layer(x)
                else:
                    h = layer(h)
            else:
                if i == 0: 
                    h = layer(x)
                else:
                    h = layer(h)
        return h
    
    def decode(self, z: torch.Tensor):
        '''
        decode features in the nth layer 
        '''
        if self.encode_only:
            raise TypeError("This is an encoder-only SAE instance")
        h = None
        for i in range(self.num_layers):
            layer = self.get_layer(FCDEF.DECODER, self.num_layers - 1 - i)
            if i == self.num_layers - 1:
                if i == 0:
                    h = layer(z)
                else:
                    h = layer(h)
            else:
                if i == 0:
                    h = layer(z)
                else:
                    h = layer(h)
        return h

    def forward(self, x: torch.Tensor):
        z = self.encode(x.view(-1, self.dim))
        return self.decode(z), z

    def fit(self, X_train, max_epoch, minibatch = True, lr = 1e-3):
        optimizer = optim.Adam(self.parameters(), lr, weight_decay = 1e-3);
        scheduler =  ReduceLROnPlateau(optimizer, mode="min", patience=10)
        if minibatch:
            for epoch in range(1, max_epoch+1):
                epoch_total_loss = 0
                for batch_index, X in enumerate(X_train):
                    if X.device.type != self.device:
                        X = X.to(self.device)
                    recon_batch, hidden_batch = self.forward(X)
                    loss = LossFunction.mse(recon_batch, X)
                    epoch_total_loss += loss.item()
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()
                    scheduler.step(loss)
                    del(X) # Reduce memory 
                    torch.cuda.empty_cache() 

                mt('Epoch: {} Average loss: {:.8f}'.format(epoch, epoch_total_loss))
                self.loss.append(epoch_total_loss)
        else:
            for epoch in range(1, max_epoch+1):
                epoch_total_loss = 0
                if X_train.device.type != self.device:
                    X_train = X_train.to(self.device)
                recon_batch, hidden_batch = self.forward(X_train)
                loss = LossFunction.mse(recon_batch, X_train)
                epoch_total_loss += loss.item()
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                scheduler.step(loss)
                torch.cuda.empty_cache() 
                mt('Epoch: {} Average loss: {:.8f}'.format(epoch, epoch_total_loss))
                self.loss.append(epoch_total_loss)
    def to(self, device:str):
        super(SAE, self).to(device)
        self.device=device 
        return self

class PositionalEncoding(nn.Module):
    def __init__(self, *,
        n_hiddens: int, 
        dropout: float, 
        max_len:int=1000
    ) -> None:
        super(PositionalEncoding, self).__init__()
        assert(n_hiddens % 2 == 0)
        self.n_hiddens = n_hiddens
        self.dropout = nn.Dropout(dropout)
        self.P = torch.zeros((1, max_len, n_hiddens))
        X = rearrange(torch.arange(max_len, dtype=torch.float32),
            '(n d) -> n d', d = 1) / torch.pow(1e4, torch.arange(
            0, n_hiddens, 2, dtype=torch.float32) / n_hiddens)
        self.P[:,:,0::2] = torch.sin(X)
        self.P[:,:,1::2] = torch.cos(X)

    def forward(self, X):
        X = X + self.P[:, :X.shape[1], :X.shape[2]].to(X.device)
        return self.dropout(X)

class FCSelfAttention(nn.Module):
    def __init__(
        self,
        *,
        dim,
        dim_head,
        n_heads
    ):
        super().__init__()
        inner_dim = dim_head * n_heads
        self.n_heads = n_heads
        self.scale = dim_head ** -0.5
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)
        self.to_out = nn.Linear(inner_dim, dim)

    def forward(self, X, mask = None):
        """ X.shape = (batch, tokens, dim)
        """
        h = self.n_heads
        q, k, v = self.to_qkv(X).chunk(3, dim = -1)
        # q.shape = (batch, head, tokens, inner_dim)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
        # attention.shape = (batch, head, tokens, tokens)
        attention = einsum('b h n d, b h m d -> b h n m', k, q) * self.scale
        if mask != None:
            if mask.shape != attention.shape:
                mask = rearrange(mask, 'n h -> n () h ()')
            attention = attention.masked_fill(mask, -torch.finfo(attention.dtype).max)
        attention = torch.softmax(attention, dim=-1)
        # out.shape = (batch, head, inner_dim, tokens)
        out = einsum('b h n e, b h e d -> b h d n', attention, v)
        # out.shape = (batch, tokens, head * inner_dim)
        out = rearrange(out, 'b h d n -> b n (h d)')
        return self.to_out(out)

class AddNorm(nn.Module):
    """Residual connection followed by layer normalization.""" 
    def __init__(self, normalized_shape, dropout, **kwargs):
        super(AddNorm, self).__init__(**kwargs) 
        self.dropout = nn.Dropout(dropout) 
        self.ln = nn.LayerNorm(normalized_shape)
    
    def forward(self, X, Y):
        return self.ln(self.dropout(Y) + X)

# reference: https://github.com/daiquocnguyen/Graph-Transformer/blob/master/UGformerV2_PyTorch/UGformerV2.py
class GraphConvolution(torch.nn.Module):
    def __init__(self, in_dim, out_dim, activation = torch.relu, bias:bool = False) -> None:
        super(GraphConvolution, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim 
        self.weight = nn.Linear(in_dim, out_dim, bias=bias)
        self.activation = activation
        self.bn = nn.BatchNorm1d(out_dim)
        self.has_bias = bias

    def forward(self, X, A):
        Z = self.weight(X)
        output = torch.mm(A, Z)
        output = self.bn(output)
        return self.activation(output)

# reference: https://github.com/daiquocnguyen/Graph-Transformer/blob/master/UGformerV2_PyTorch/UGformerV2.py    
class FullyConnectedGraphTransformer(nn.Module):
    def __init__(self, feature_dim,
                       ff_hidden_size,
                       n_self_att_layers,
                       dropout,
                       n_GNN_layers,
                       n_head, device = "cuda"
                       ) -> None:
        super(FullyConnectedGraphTransformer, self).__init__()
        self.feature_dim = feature_dim
        self.ff_hidden_size = ff_hidden_size
        self.n_self_att_layers = n_self_att_layers
        self.n_GNN_layers = n_GNN_layers
        self.n_head = n_head
        self.GNN_layers = nn.ModuleList()
        self.selfAtt_layers = nn.ModuleList()
        self.dropouts = nn.ModuleList()
        self.node_embedding = nn.Linear(feature_dim * n_GNN_layers, feature_dim, bias = True)
        self.device = device
        for _ in range(self.n_GNN_layers):
            encoder_layer = nn.TransformerEncoderLayer(self.feature_dim, 
                nhead=self.n_head, 
                dim_feedforward=self.ff_hidden_size,
                dropout=0.5)
            self.selfAtt_layers.append(
                nn.TransformerEncoder(encoder_layer=encoder_layer, num_layers=self.n_self_att_layers).to(device))
            self.GNN_layers.append(GraphConvolution(self.feature_dim, self.feature_dim, torch.relu, True).to(device))
            self.dropouts.append(nn.Dropout(dropout))
    
    def reset_parameters(self):
        for i in self.selfAtt_layers:
            i.reset_parameters()
        self.prediction.reset_parameters()
        for i in self.dropouts:
            i.reset_parameters()

    def forward(self, X, A):
        Xs = []
        for i in range(self.n_GNN_layers):
            # self attention over all nodes
            X = X.unsqueeze(1)
            X = self.selfAtt_layers[i](X)
            X = X.squeeze(1)
            X = self.GNN_layers[i](X, A)
            Xs.append(X)
        X = torch.hstack(Xs)
        X = self.node_embedding(X)
        return X

    def to(self, device:str):
        super(FullyConnectedGraphTransformer, self).to(device)
        self.device=device 
        return self


import multiprocessing
import os
import scirpy as ir
import scanpy as sc
import scipy
import numpy as np
import pandas as pd
from scipy import sparse
from collections import Counter
import matplotlib.pyplot as plt
from typing import Callable, List, Optional, Union, Iterable
from functools import partial
from pathlib import Path
from ..utils._logger import mt
from ..utils._decorators import typed
from ..utils._definitions import (
    TRA_DEFINITION_ORIG,
    TRAB_DEFINITION_ORIG,
    TRB_DEFINITION_ORIG,
    TRA_DEFINITION,
    TRAB_DEFINITION,
    TRB_DEFINITION,
)
from ..utils._logger import mt, Colors, get_tqdm


MODULE_PATH = Path(__file__).parent


plt.ioff()

"""
Reference human T/B cell annotation data for singleR. Originated from
Monaco, et al. (2019). RNA-Seq Signatures Normalized by mRNA Abundance Allow Absolute Deconvolution of Human Immune Cell Types. 
Cell Reports 26, 1627-1640.e7. 10.1016/j.celrep.2019.01.041.
The reference data is subsetted to only contain T or B cells, removing other immune cell types.

:attr:`Tonly`: T cell only reference data
:attr:`TBonly`: T/B cell reference data
"""
HSAP_REF_DATA = {
    "Tonly": MODULE_PATH / "./data/refdata/human/reft_name.Rdata",
    "TBonly": MODULE_PATH / "./data/refdata/human/reft_name_tb.Rdata"
}



MMUS_REF_DATA = {
    "Tonly_no_thymus": MODULE_PATH / "./data/refdata/mouse/reft_name_mouse_nothymus.rds",
    "Tonly_with_thymus":  MODULE_PATH / "./data/refdata/mouse/reft_name_mouse.rds"
}
"""
Reference muose T/B cell annotation data for singleR
Monaco, et al. (2019). RNA-Seq Signatures Normalized by mRNA Abundance Allow Absolute Deconvolution of Human Immune Cell Types. 
Cell Reports 26, 1627-1640.e7. 10.1016/j.celrep.2019.01.041.
The reference data is subsetted to only contain T cells, removing other immune cell types. For pro-T and pre-T cells, 
use `Tonly_with_thymus`. For mature T cells, use `Tonly_no_thymus`.

:attr:`Tonly_no_thymus`: T cell only reference data without thymus derived T cells
:attr:`Tonly_with_thymus`: T cell only reference data with thymus derived T cells
"""


class VDJPreprocessingV1Human:
    def __init__(
        self,
        *,
        cellranger_gex_output_path: str,
        cellranger_vdj_output_path: str,
        output_path: str,
        check_existing_files: bool = False,
        check_valid_vdj: bool = True,
        vdj_all_contig: bool = False,
        cellranger_gex_output_path_opt: Optional[str] = None,
        cellranger_vdj_output_path_opt: Optional[Union[str, Iterable[str]]] = None,
        sample_name: Optional[str] = None,
        study_name: Optional[str] = None
    ):
        """
        Preprocess the output of cellranger vdj and cellranger gex.

        :param cellranger_gex_output_path: Path to the cellranger gex output folder.
        :param cellranger_vdj_output_path: Path to the cellranger vdj output folder.
        :param output_path: Path to the output folder.
        :param check_existing_files: If True, check if the output files already exist and skip the computation.
        :param check_valid_vdj: If True, check if the vdj data is valid.
        :param vdj_all_contig: If True, use the all_contig_annotations.json file instead of filtered_contig_annotations.csv.
        :param cellranger_gex_output_path_opt: Path to the cellranger gex output folder for the optional sample.
        :param cellranger_vdj_output_path_opt: Path to the cellranger vdj output folder for the optional sample.
        :param sample_name: Name of the sample.
        :param study_name: Name of the study.

        :example:
        >>> # Preprocess the output of Human sample cellranger vdj and cellranger gex.
        >>> import scatlasvae
        >>> pp = scatlasvae.pp.VDJPreprocessingV1Human(
        ...     cellranger_gex_output_path = "./cellranger_gex_output",
        ...     cellranger_vdj_output_path = "./cellranger_vdj_output",
        ...     output_path = "./output",
        ... )
        >>> pp.process(
        ...     r_path = "/opt/anaconda3/envs/r403/bin/Rscript",
        ...     ref_data_path = scatlasvae.pp.HSAP_REF_DATA["Tonly"]
        ... )
        """
        self.filtered_10x_feature_bc_matrix_path = os.path.join(
            os.path.abspath(cellranger_gex_output_path), "filtered_feature_bc_matrix"
        )
        self.filtered_10x_feature_bc_matrix_path_opt = None
        if cellranger_gex_output_path_opt is not None:
            self.filtered_10x_feature_bc_matrix_path_opt = os.path.join(
                os.path.abspath(cellranger_gex_output_path_opt), "filtered_feature_bc_matrix"
            )

        self.raw_10x_feature_bc_matrix_path = os.path.join(
            os.path.abspath(cellranger_gex_output_path), "raw_feature_bc_matrix"
        )
        self.output_path = os.path.abspath(output_path)
        self.vdj_all_contig = vdj_all_contig
        if not os.path.exists(self.output_path):
            os.mkdir(self.output_path)

        self.vdj_data_path = os.path.join(cellranger_vdj_output_path, 'filtered_contig_annotations.csv')
        self.vdj_data_path_all_contig = os.path.join(cellranger_vdj_output_path, 'all_contig_annotations.json')
        
        self.check_valid_vdj = check_valid_vdj

        if cellranger_vdj_output_path_opt:
            if isinstance(cellranger_vdj_output_path_opt, str):
                self.vdj_data_path_opt = os.path.join(cellranger_vdj_output_path_opt, 'filtered_contig_annotations.csv')
                self.vdj_data_path_opt_all_contig = os.path.join(cellranger_vdj_output_path, 'all_contig_annotations.json')

            elif isinstance(cellranger_vdj_output_path_opt, Iterable):
                self.vdj_data_path_opt = list(map(lambda x: os.path.join(x, 'filtered_contig_annotations.csv'), cellranger_vdj_output_path_opt))
                self.vdj_data_path_opt_all_contig = list(map(lambda x: os.path.join(x, 'all_contig_annotations.json'), cellranger_vdj_output_path_opt))
        else: 
            self.vdj_data_path_opt = None
            os.system(f"cp {self.vdj_data_path_all_contig} {os.path.join(self.output_path, 'all_contig_annotations.json')}")

        self.sample_name = sample_name
        self.study_name = study_name
        self.statistics = {}


    def process(self, 
        r_path: str = "/opt/anaconda3/envs/r403/bin/Rscript", 
        ref_data_path: str = HSAP_REF_DATA["Tonly"] 
    ):
        """
        Preprocess the output of cellranger vdj and cellranger gex.
        
        :param r_path: Path to the Rscript executable. 
        :param ref_data_path: Path to the reference data. Available references are defined in `scatlasvae.pp.HSAP_REF_DATA` and `scatlasvae.pp.MMUS_REF_DATA`.
        """
        self._initialize_filtered_adata()
        self._initialize_vdj_data()
        self._merge_adata_with_vdj()
        if self._filtered_adata.shape[0] == 0:
            print("No cells with IR information found. Exiting...")
            return
        self._cell_annotation(r_path = r_path, ref_data_path = ref_data_path)
        self._tcr_preprocessing()
        self._gex_preprocessing()
        
        
    def _check_output_path(self):
        """
        Check if the output files already exist. If not, create the output folder.

        :return: True if the output files already exist, False otherwise.
        """
        if os.path.exists(os.path.join(self.output_path, "results_raw.h5ad")) and os.path.exists(os.path.join(self.output_path, "results_single_chain_raw.h5ad")):
            self._filtered_adata = sc.read_h5ad(os.path.join(self.output_path, "results_raw.h5ad"))
            self._filtered_adata_single_chain = sc.read_h5ad(os.path.join(self.output_path, "results_single_chain_raw.h5ad"))
            return True 
        else:
            return False

    def _initialize_filtered_adata(self):
        """
        Initialize the filtered adata object.

        :return: None
        """
        self._filtered_adata = sc.read_10x_mtx(self.filtered_10x_feature_bc_matrix_path)

        if self.filtered_10x_feature_bc_matrix_path_opt is not None:
            _filtered_adata_opt = sc.read_10x_mtx(self.filtered_10x_feature_bc_matrix_path_opt)
            barcode_1 = list(self._filtered_adata.obs.index)
            barcode_2 = list(_filtered_adata_opt.obs.index)
            all_barcodes = sorted(list(set(barcode_1).union(set(barcode_2))))
            all_barcodes_index = dict(zip(all_barcodes, list(range(len(all_barcodes)))))
            new_X = np.zeros((len(all_barcodes), self._filtered_adata.shape[1]))
            indices_1 = list(map(lambda x: all_barcodes_index[x], barcode_1))
            indices_2 = list(map(lambda x: all_barcodes_index[x], barcode_2))
            new_X[indices_1] += self._filtered_adata.X.toarray()
            new_X[indices_2] += _filtered_adata_opt.X.toarray()
            self._filtered_adata = sc.AnnData(X = sparse.csr_matrix(new_X), obs = pd.DataFrame(index=all_barcodes), var = self._filtered_adata.var)

        self._raw_filtered_adata = self._filtered_adata.copy()
        self._filtered_adata.var_names_make_unique()
        self.statistics["filtered_adata_original"] = {
            "cell_number": self._filtered_adata.shape[0],
            "gene_number": self._filtered_adata.shape[1]
        }

    def _initialize_vdj_data(self):
        """
        Initialize the vdj data.

        :return: None
        """
        if self.vdj_all_contig:
            self._vdj = ir.io.read_10x_vdj(self.vdj_data_path_all_contig)
        else:
            self._vdj = ir.io.read_10x_vdj(self.vdj_data_path)
        if isinstance(self.vdj_data_path_opt, str):
            self._vdj_opt = ir.io.read_10x_vdj(self.vdj_data_path_opt)
        elif isinstance(self.vdj_data_path_opt, Iterable):
            self._vdj_opt = list(map(ir.io.read_10x_vdj, self.vdj_data_path_opt))
        else: 
            self._vdj_opt = None 

        if self.vdj_data_path_opt is not None:
            if sorted(Counter(list(map(lambda z: z[:2], filter(lambda x: type(x) == str, self._vdj.obs['IR_VJ_1_v_call'])))).items(), key=lambda x: -x[1])[0][0] == 'TR':
                os.system(f"cp {self.vdj_data_path_all_contig} {os.path.join(self.output_path, 'all_contig_annotations_TCR.json')}")
            else:
                os.system(f"cp {self.vdj_data_path_all_contig} {os.path.join(self.output_path, 'all_contig_annotations_BCR.json')}")


    def _merge_adata_with_vdj(self):
        """
        Merge the filtered adata with the vdj data.

        :return: None
        """
        ir.pp.merge_with_ir(self._filtered_adata, self._vdj)
        if isinstance(self._vdj_opt, sc.AnnData):
            self._filtered_adata.obs['is_cell'] = 'None'
            self._filtered_adata.obs['high_confidence'] = 'None'
            self._vdj_opt.obs['is_cell'] = 'None'
            self._vdj_opt.obs['high_confidence'] = 'None'
            ir.pp.merge_airr_chains(self._filtered_adata, self._vdj_opt)
        elif isinstance(self._vdj_opt, Iterable):
            for i in self._vdj_opt:
                self._filtered_adata.obs['is_cell'] = 'None'
                self._filtered_adata.obs['high_confidence'] = 'None'
                i.obs['is_cell'] = 'None'
                i.obs['high_confidence'] = 'None'
                ir.pp.merge_airr_chains(self._filtered_adata, i)
        ir.tl.chain_qc(self._filtered_adata)
        if self.check_valid_vdj:
            self._filtered_adata = self._filtered_adata[
                list(map(lambda x: x in ["single pair", "extra VJ", "extra VDJ"], self._filtered_adata.obs["chain_pairing"]))
            ].copy()
        self._filtered_adata_single_chain = self._filtered_adata[self._filtered_adata.obs["chain_pairing"] == "single pair"]
        self._filtered_adata.write_h5ad(os.path.join(self.output_path, "results_raw.h5ad"))
        self._filtered_adata_single_chain.write_h5ad(os.path.join(self.output_path, "results_single_chain_raw.h5ad"))

    def _gex_preprocessing(self):
        pipeline = lambda adata: [f(adata) for f in [
            partial(sc.pp.normalize_total, target_sum = 1e6),
            sc.pp.log1p,
            partial(
                sc.pp.highly_variable_genes, 
                min_mean=0.0125, 
                max_mean=np.max(adata.X.toarray()), 
                min_disp=0.5
            ),
            partial(sc.pp.pca, svd_solver="arpack"),
            sc.pp.neighbors,
            partial(sc.tl.tsne, n_jobs=int(multiprocessing.cpu_count() / 2)),
            partial(sc.tl.umap),
            sc.tl.leiden
        ]]
        self._filtered_adata_preprocessed = self._filtered_adata.copy()
        self._filtered_adata_single_chain_preprocessed = self._filtered_adata_single_chain.copy()
        pipeline(self._filtered_adata_preprocessed)
        pipeline(self._filtered_adata_single_chain_preprocessed)
        self._filtered_adata_preprocessed.write_h5ad(os.path.join(self.output_path, "results_preprocessed.h5ad"))
        self._filtered_adata_single_chain_preprocessed.write_h5ad(os.path.join(self.output_path, "results_single_chain_preprocessed.h5ad"))

    def _tcr_preprocessing(self):
        pipeline = lambda adata: [f(adata) for f in [
            ir.pp.ir_dist,
            partial(ir.tl.define_clonotypes, receptor_arms="all", dual_ir="any"),
            partial(ir.pp.ir_dist,  metric='hamming', sequence='aa', cutoff=2, n_jobs=4),
            partial(ir.tl.define_clonotype_clusters, partitions="connected",sequence='aa', metric='hamming'),
            partial(ir.tl.clonotype_network, min_cells=1, sequence='aa', metric='hamming')
        ]]
        pipeline(self._filtered_adata)
        pipeline(self._filtered_adata_single_chain)

    def _update_annotated(self, _filtered_adata_annotated: sc.AnnData):
        """
        This method adapts `20210820UpdatePrediction.py`
        """
        ## choose T helper cells
        Th_cells = _filtered_adata_annotated[_filtered_adata_annotated.obs.predictions.str.contains('Th|Follicular'),:]
        ## CD8A expression
        if "CD8A" in Th_cells.var.index:
            Th_cells_CD8A_np_matrix = Th_cells[:,['CD8A']].X.todense()
        else: 
            Th_cells_CD8A_np_matrix = self._raw_filtered_adata[Th_cells.obs.index, ['CD8A']].X.todense()
        ## Flatten
        Th_cells_CD8A_expression = [y for x in Th_cells_CD8A_np_matrix.tolist() for y in x]
        Th_cells_CD8A = pd.DataFrame({"barcode":Th_cells.obs.index.to_list(),
                                    'expression':Th_cells_CD8A_expression})
        ## choose cells whose expression of CD8A is larger than 0.
        Th_cells_CD8A_filter_barcodes = Th_cells_CD8A[Th_cells_CD8A['expression']>0]['barcode'].to_list()
        ## choose CD8 cells
        CD8_cells = _filtered_adata_annotated[_filtered_adata_annotated.obs.predictions.str.contains('CD8'),:]
        ## CD4 expression
        if "CD4" in CD8_cells.var.index:
            CD8_cells_np_matrix = CD8_cells[:,['CD4']].X.todense()
        else: 
            CD8_cells_np_matrix = self._raw_filtered_adata[CD8_cells.obs.index, ['CD4']].X.todense()
        ## Flatten
        CD8_cells_CD4_expression = [y for x in CD8_cells_np_matrix.tolist() for y in x]
        CD8_cells_CD4 = pd.DataFrame({"barcode":CD8_cells.obs.index.to_list(),
                                    'expression':CD8_cells_CD4_expression})
        ## choose cells whose expression of CD4 is larger than 0.
        CD8_cells_CD4_filter_barcodes = CD8_cells_CD4[CD8_cells_CD4['expression']>0]['barcode'].to_list()

        ### update predictions
        predictions = _filtered_adata_annotated.obs.predictions
        ## add Upredictive category before update

        # TODO: Deal with the predictions
        try:
            predictions = predictions.cat.add_categories(['Unpredictive'])
        except:
            pass 

        for i in (CD8_cells_CD4_filter_barcodes + Th_cells_CD8A_filter_barcodes):
            predictions[i] = 'Unpredictive'
        _filtered_adata_annotated.obs.predictions = predictions

    def _cell_annotation(self, r_path: str, ref_data_path: str = HSAP_REF_DATA["Tonly"]):
        r_script = (
            """
            library(Seurat)
            library(DoubletFinder)
            library(dplyr)
            library(SingleR)
            library(zellkonverter)
            library(SeuratDisk)
            RhpcBLASctl::blas_set_num_threads(6)
            """
            f"rawdata = Read10X('{self.filtered_10x_feature_bc_matrix_path}')\n"
            """
            pdf(NULL)
            seurat_object = CreateSeuratObject(rawdata, min.cells = 3, min.features = 200)
            seurat_object[["percent.mt"]] = PercentageFeatureSet(seurat_object,pattern = '^MT-')
            seurat_object = subset(seurat_object, subset = nFeature_RNA > 200 & percent.mt < 20)
            seurat_object = NormalizeData(seurat_object)
            seurat_object = FindVariableFeatures(seurat_object,selection.method = 'vst',nfeatures = 2000)
            seurat_object = ScaleData(seurat_object,verbose = FALSE, features = rownames(seurat_object))
            seurat_object = RunPCA(seurat_object, verbose = FALSE)
            sweep.res.list = paramSweep_v3(seurat_object, PCs = 1:50)
            sweep.stats = summarizeSweep(sweep.res.list, GT = FALSE)
            bcmvn = find.pK(sweep.stats)
            pk_bcmvn = as.numeric(as.vector(bcmvn$pK[which.max(bcmvn$BCmetric)]))
            cellnumber = ncol(seurat_object)
            multiplet_rate = 0
            if (cellnumber <= 500) {
            multiplet_rate = 0.004
            } else if ( cellnumber > 500 & cellnumber <= 1000) {
            multiplet_rate = 0.008
            } else if ( cellnumber > 100 & cellnumber <= 2000) {
            multiplet_rate = 0.016
            } else if ( cellnumber > 2000 & cellnumber <= 3000) {
            multiplet_rate = 0.023
            } else if ( cellnumber > 3000 & cellnumber <= 4000) {
            multiplet_rate = 0.031
            } else if ( cellnumber > 4000 & cellnumber <= 5000) { 
            multiplet_rate = 0.039
            } else if ( cellnumber > 5000 & cellnumber <= 6000) {
            multiplet_rate = 0.046
            } else if ( cellnumber > 6000 & cellnumber <= 7000) {
            multiplet_rate = 0.054
            } else if ( cellnumber > 7000 & cellnumber <= 8000) {
            multiplet_rate = 0.061
            } else if ( cellnumber > 8000 & cellnumber <= 9000) {
            multiplet_rate = 0.069
            } else {
            multiplet_rate = 0.076
            }

            annotations = seurat_object@meta.data$seurat_clusters
            homotypic.prop = modelHomotypic(annotations)  
            nExp_poi = round(multiplet_rate*cellnumber)  
            nExp_poi.adj = round(nExp_poi*(1-homotypic.prop))

            seurat_object_filterDoublet = doubletFinder_v3(
                seurat_object, 
                PCs = 1:50, pN = 0.25, 
                pK = pk_bcmvn, nExp = nExp_poi, 
                reuse.pANN = FALSE
            )
            seurat_object_filterDoublet = subset(seurat_object_filterDoublet,
                cells = rownames(seurat_object_filterDoublet[[paste("DF.classifications_0.25",pk_bcmvn,nExp_poi,sep = '_')]])
            )
            """
            f"load('{ref_data_path}')\n"
            """
            sce <- as.SingleCellExperiment(seurat_object_filterDoublet)
            predictions <- SingleR(test = sce, ref = reft, labels = reft$label.fine)
            sce[["predictions"]] <- predictions$labels
            """ 
            f"SeuratDisk::SaveH5Seurat(as.Seurat(sce), '{os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}')\n"
            f"""SeuratDisk::Convert("{os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}", dest="h5ad")\n"""
        )
        with open(os.path.join(self.output_path, "rscript.R"), "w+") as f:
            f.write(r_script)
        if os.path.exists(os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')):
            os.system(f"rm {os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}")
        os.system(f'{r_path} {os.path.join(self.output_path, "rscript.R")}')
        os.system(f"rm {os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}")
        _filtered_adata_annotated = sc.read_h5ad(os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5ad'))
        self._update_annotated(_filtered_adata_annotated)
        self._filtered_adata = self._filtered_adata[list(map(lambda x: x in  _filtered_adata_annotated.obs.index, self._filtered_adata.obs.index))].copy()
        self._filtered_adata.obs["predictions"] = _filtered_adata_annotated.obs.loc[self._filtered_adata.obs.index, "predictions"]
        self._filtered_adata_single_chain = self._filtered_adata_single_chain[list(map(lambda x: x in _filtered_adata_annotated.obs.index, self._filtered_adata_single_chain.obs.index))].copy()
        self._filtered_adata_single_chain.obs["predictions"] = _filtered_adata_annotated.obs.loc[self._filtered_adata_single_chain.obs.index, "predictions"]
    

class VDJPreprocessingV1Mouse(VDJPreprocessingV1Human):
    def process(self, r_path: str = "/opt/anaconda3/envs/r403/bin/Rscript", ref_data_path: str = MMUS_REF_DATA["Tonly_no_thymus"] ):
        self._initialize_filtered_adata()
        self._initialize_vdj_data()
        self._merge_adata_with_vdj()
        self._cell_annotation(r_path = r_path, ref_data_path = ref_data_path)
        self._tcr_preprocessing()
        self._gex_preprocessing()

    def _update_annotated(self, _filtered_adata_annotated: sc.AnnData):
        """
        This method adapts `20210820UpdatePrediction.py`
        Mouse version adapting human genes
        """
        ## choose CD4 T cells
        Th_cells = _filtered_adata_annotated[
            np.array(_filtered_adata_annotated.obs.predictions_new.str.contains('CD4')) |
            np.array(_filtered_adata_annotated.obs.predictions_new.str.contains('T.4'))
        ,:]
        ## CD8A expression
        if "Cd8a" in Th_cells.var.index:
            Th_cells_CD8A_np_matrix = Th_cells[:,['Cd8a']].X.todense()
        else: 
            Th_cells_CD8A_np_matrix = self._raw_filtered_adata[Th_cells.obs.index, ['Cd8a']].X.todense()
        ## Flatten
        Th_cells_CD8A_expression = [y for x in Th_cells_CD8A_np_matrix.tolist() for y in x]
        Th_cells_CD8A = pd.DataFrame({"barcode":Th_cells.obs.index.to_list(),
                                    'expression':Th_cells_CD8A_expression})
        ## choose cells whose expression of CD8A is larger than 0.
        Th_cells_CD8A_filter_barcodes = Th_cells_CD8A[Th_cells_CD8A['expression']>0]['barcode'].to_list()
        ## choose CD8 cells
        CD8_cells = _filtered_adata_annotated[
            np.array(_filtered_adata_annotated.obs.predictions_new.str.contains('CD8')) | 
            np.array(_filtered_adata_annotated.obs.predictions_new.str.contains('T.8'))
        ,:]
        ## CD4 expression
        if "Cd4" in CD8_cells.var.index:
            CD8_cells_np_matrix = CD8_cells[:,['Cd4']].X.todense()
        else: 
            CD8_cells_np_matrix = self._raw_filtered_adata[CD8_cells.obs.index, ['Cd4']].X.todense()
        ## Flatten
        CD8_cells_CD4_expression = [y for x in CD8_cells_np_matrix.tolist() for y in x]
        CD8_cells_CD4 = pd.DataFrame({"barcode":CD8_cells.obs.index.to_list(),
                                    'expression':CD8_cells_CD4_expression})
        ## choose cells whose expression of CD4 is larger than 0.
        CD8_cells_CD4_filter_barcodes = CD8_cells_CD4[CD8_cells_CD4['expression']>0]['barcode'].to_list()

        ### update predictions
        predictions_new = _filtered_adata_annotated.obs.predictions_new
        ## add Upredictive category before update

        # TODO: Deal with the predictions
        try:
            predictions_new = predictions_new.cat.add_categories(['Unpredictive'])
        except:
            pass 

        for i in (CD8_cells_CD4_filter_barcodes + Th_cells_CD8A_filter_barcodes):
            predictions_new[i] = 'Unpredictive'
        _filtered_adata_annotated.obs.predictions_new = predictions_new

        
    def _cell_annotation(self, r_path: str, ref_data_path: str = MMUS_REF_DATA["Tonly_no_thymus"]):
        r_script = (
            """
            library(Seurat)
            library(DoubletFinder)
            library(dplyr)
            library(SingleR)
            library(zellkonverter)
            library(SeuratDisk)
            RhpcBLASctl::blas_set_num_threads(6)
            """
            f"rawdata = Read10X('{self.filtered_10x_feature_bc_matrix_path}')\n"
            """
            pdf(NULL)
            seurat_object = CreateSeuratObject(rawdata, min.cells = 3, min.features = 200)
            seurat_object[["percent.mt"]] = PercentageFeatureSet(seurat_object,pattern = '^MT-')
            seurat_object = subset(seurat_object, subset = nFeature_RNA > 200 & percent.mt < 20)
            seurat_object = NormalizeData(seurat_object)
            seurat_object = FindVariableFeatures(seurat_object,selection.method = 'vst',nfeatures = 2000)
            seurat_object = ScaleData(seurat_object,verbose = FALSE, features = rownames(seurat_object))
            seurat_object = RunPCA(seurat_object, verbose = FALSE)
            sweep.res.list = paramSweep_v3(seurat_object, PCs = 1:50)
            sweep.stats = summarizeSweep(sweep.res.list, GT = FALSE)
            bcmvn = find.pK(sweep.stats)
            pk_bcmvn = as.numeric(as.vector(bcmvn$pK[which.max(bcmvn$BCmetric)]))
            cellnumber = ncol(seurat_object)
            multiplet_rate = 0
            if (cellnumber <= 500) {
            multiplet_rate = 0.004
            } else if ( cellnumber > 500 & cellnumber <= 1000) {
            multiplet_rate = 0.008
            } else if ( cellnumber > 100 & cellnumber <= 2000) {
            multiplet_rate = 0.016
            } else if ( cellnumber > 2000 & cellnumber <= 3000) {
            multiplet_rate = 0.023
            } else if ( cellnumber > 3000 & cellnumber <= 4000) {
            multiplet_rate = 0.031
            } else if ( cellnumber > 4000 & cellnumber <= 5000) { 
            multiplet_rate = 0.039
            } else if ( cellnumber > 5000 & cellnumber <= 6000) {
            multiplet_rate = 0.046
            } else if ( cellnumber > 6000 & cellnumber <= 7000) {
            multiplet_rate = 0.054
            } else if ( cellnumber > 7000 & cellnumber <= 8000) {
            multiplet_rate = 0.061
            } else if ( cellnumber > 8000 & cellnumber <= 9000) {
            multiplet_rate = 0.069
            } else {
            multiplet_rate = 0.076
            }

            annotations = seurat_object@meta.data$seurat_clusters
            homotypic.prop = modelHomotypic(annotations)  
            nExp_poi = round(multiplet_rate*cellnumber)  
            nExp_poi.adj = round(nExp_poi*(1-homotypic.prop))

            seurat_object_filterDoublet = doubletFinder_v3(
                seurat_object, 
                PCs = 1:50, pN = 0.25, 
                pK = pk_bcmvn, nExp = nExp_poi, 
                reuse.pANN = FALSE
            )
            seurat_object_filterDoublet = subset(seurat_object_filterDoublet,
                cells = rownames(seurat_object_filterDoublet[[paste("DF.classifications_0.25",pk_bcmvn,nExp_poi,sep = '_')]])
            )
            """
            f"reft <- readRDS('{ref_data_path}')\n"
            """
            sce <- as.SingleCellExperiment(seurat_object_filterDoublet)
            predictions <- SingleR(test = sce, ref = reft, labels = reft$label.new)
            sce[["predictions_new"]] <- predictions$labels
            predictions <- SingleR(test = sce, ref = reft, labels = reft$label.fine)
            sce[["predictions_fine"]] <- predictions$labels
            """ 
            f"SeuratDisk::SaveH5Seurat(as.Seurat(sce), '{os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}')\n"
            f"""SeuratDisk::Convert("{os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}", dest="h5ad")\n"""
        )
        with open(os.path.join(self.output_path, "rscript.R"), "w+") as f:
            f.write(r_script)
        if os.path.exists(os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')):
            os.system(f"rm {os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}")
        os.system(f'{r_path} {os.path.join(self.output_path, "rscript.R")}')
        os.system(f"rm {os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5seurat')}")
        _filtered_adata_annotated = sc.read_h5ad(os.path.join(self.output_path, 'filterDoublet_prediction_sce.h5ad'))
        self._update_annotated(_filtered_adata_annotated)

        self._filtered_adata = self._filtered_adata[list(map(lambda x: x in  _filtered_adata_annotated.obs.index, self._filtered_adata.obs.index))].copy()
        self._filtered_adata.obs["predictions_fine"] = _filtered_adata_annotated.obs.loc[self._filtered_adata.obs.index, "predictions_fine"]
        self._filtered_adata_single_chain = self._filtered_adata_single_chain[list(map(lambda x: x in _filtered_adata_annotated.obs.index, self._filtered_adata_single_chain.obs.index))].copy()
        self._filtered_adata_single_chain.obs["predictions_fine"] = _filtered_adata_annotated.obs.loc[self._filtered_adata_single_chain.obs.index, "predictions_fine"]

        self._filtered_adata = self._filtered_adata[list(map(lambda x: x in  _filtered_adata_annotated.obs.index, self._filtered_adata.obs.index))].copy()
        self._filtered_adata.obs["predictions_new"] = _filtered_adata_annotated.obs.loc[self._filtered_adata.obs.index, "predictions_new"]
        self._filtered_adata_single_chain = self._filtered_adata_single_chain[list(map(lambda x: x in _filtered_adata_annotated.obs.index, self._filtered_adata_single_chain.obs.index))].copy()
        self._filtered_adata_single_chain.obs["predictions_new"] = _filtered_adata_annotated.obs.loc[self._filtered_adata_single_chain.obs.index, "predictions_new"]


@typed({
    'gex_adata': sc.AnnData, 
    'gex_embedding_key': str,
    'tcr_embedding_key': str,
    'joint_embedding_key': str,
})
def update_anndata(
    gex_adata: sc.AnnData,
    gex_embedding_key: str = 'X_gex',
    tcr_embedding_key: str = 'X_tcr',
    joint_embedding_key: str = 'X_gex_tcr',
):
    """
    Update the adata with the embedding keys

    .. note::
        This method modifies the `gex_adata` inplace. 
        added columns in .obs: `tcr`, `CDR3a`, `CDR3b`, `TRAV`, `TRAJ`, `TRBV`, `TRBJ`

    .. note::
        `"tcr"` should be in `gex_adata.obs.columns`.


    :param gex_adata: AnnData object
    :param gex_embedding_key: embedding key for gex
    :param tcr_embedding_key: embedding key for tcr
    :param joint_embedding_key: embedding key for joint
    """
    gex_adata.uns['embedding_keys'] = {
        "gex": gex_embedding_key,
        "tcr": tcr_embedding_key,
        "joint": joint_embedding_key,
    }
    for i,j in zip([
        'IR_VJ_1_junction_aa',
        'IR_VDJ_1_junction_aa',
        'IR_VJ_1_v_call',
        'IR_VJ_1_j_call',
        'IR_VDJ_1_v_call',
        'IR_VDJ_1_j_call'
    ], ['CDR3a', 
        'CDR3b', 
        'TRAV', 
        'TRAJ', 
        'TRBV', 
        'TRBJ'
    ]):
        if not (i in gex_adata.obs.columns or j in gex_adata.obs.columns):
            raise ValueError(f"{i} or {j} is not in adata.obs.columns")
        if i in gex_adata.obs.columns:
            gex_adata.obs[j] = gex_adata.obs[i]

    if not 'individual' in gex_adata.obs.columns:
        raise ValueError("individual is not in adata.obs.columns.")
    mt("TCRDeepInsight: initializing dataset")
    mt("TCRDeepInsight: adding 'tcr' to adata.obs")
    gex_adata.obs['tcr'] = None
    gex_adata.obs.iloc[:, list(gex_adata.obs.columns).index("tcr")] = list(map(lambda x: '='.join(x), gex_adata.obs.loc[:,TRAB_DEFINITION + ['individual']].to_numpy()))

@typed({
    'gex_adata': sc.AnnData,
    'gex_embedding_key': str,
    'agg_index': pd.DataFrame,
})
def aggregated_gex_embedding_by_tcr(
        gex_adata: sc.AnnData,
        gex_embedding_key: str,
        agg_index: pd.DataFrame,
    ):
    """
    Aggregate GEX embedding by TCR

    :param gex_adata: AnnData object
    :param agg_index: Aggregated index

    :return: aggregated GEX embedding
    """
    all_gex_embedding = []
    for i in agg_index['index']:
        all_gex_embedding.append(
            gex_adata.obsm[gex_embedding_key][i].mean(0)
        )
    all_gex_embedding = np.vstack(all_gex_embedding)
    return all_gex_embedding

def annotate_t_cell_cd4_cd8(adata: sc.AnnData, use_rep='X_gex'):
    adata.obs["cd4_cd8_class"] = None
    adata.obs.loc[
        ((adata.X[:,list(adata.var_names).index("CD8A")].toarray()  == 0) & \
         (adata.X[:,list(adata.var_names).index("CD8B")].toarray()  == 0) & \
         (adata.X[:,list(adata.var_names).index("CD4")].toarray() > 0) ).flatten()
    ,"cd4_cd8_class"] = "CD4"
    adata.obs.loc[
        (((adata.X[:,list(adata.var_names).index("CD8A")].toarray() > 0) | \
        (adata.X[:,list(adata.var_names).index("CD8B")].toarray() > 0)) & \
        (adata.X[:,list(adata.var_names).index("CD4")].toarray() == 0) ).flatten()
    ,"cd4_cd8_class"] = "CD8"
    adata.obs.loc[
        ((adata.X[:,list(adata.var_names).index("CD8A")].toarray() == 0) & \
         (adata.X[:,list(adata.var_names).index("CD8B")].toarray() == 0) & \
        (adata.X[:,list(adata.var_names).index("CD4")].toarray() == 0) ).flatten()
    ,"cd4_cd8_class"] = "double_negative"
    adata.obs.loc[
        (((adata.X[:,list(adata.var_names).index("CD8A")].toarray() > 0) | \
        (adata.X[:,list(adata.var_names).index("CD8B")].toarray() > 0)) & \
        (adata.X[:,list(adata.var_names).index("CD4")].toarray() > 0) ).flatten()
    ,"cd4_cd8_class"] = "double_positive"
    from sklearn.neighbors import KNeighborsClassifier
    nn = KNeighborsClassifier(n_neighbors=13)
    nn.fit(adata[list(map(lambda x: 
        x in ["CD8","CD4"], adata.obs['cd4_cd8_class']))].obsm[use_rep], 
        adata[list(map(lambda x: x in ["CD8","CD4"], adata.obs['cd4_cd8_class']))].obs["cd4_cd8_class"]
    )
    Y = nn.predict(adata.obsm[use_rep])
    adata.obs["cd4_cd8_annotation"] = Y

def subset_adata_by_genes_fill_zeros(
    adata: sc.AnnData, 
    gene_list: List[str]
):
    """
    Subset adata by genes and fill zeros for missing genes

    :param adata: AnnData object
    :param gene_list: List of genes to subset

    :return: Subsetted AnnData object
    """
    low_memory = False
    if adata.shape[0] > 100000:
        low_memory = True
    X = np.zeros((adata.shape[0], len(gene_list)), dtype=np.int32)
    var = list(adata.var.index)
    pbar = get_tqdm()(range(len(gene_list)), desc='Subset by genes', bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}')
    original_X = adata.X 

    if not low_memory and scipy.sparse.issparse(original_X):
        original_X = original_X.toarray()

    for i,j in enumerate(gene_list):
        if j in var:
            if low_memory:
                X[:,i] = original_X[:, var.index(j)].toarray().flatten()
            else:
                X[:,i] = original_X[:, var.index(j)].flatten()
        pbar.update(1)
    pbar.close()
    result_adata = sc.AnnData(
        X = scipy.sparse.csr_matrix(X),
        obs = adata.obs.copy(),
        var = pd.DataFrame(
            index = gene_list
        ),
        uns = adata.uns.copy(),
        obsm = adata.obsm.copy()
    )
    return result_adata


import torch
import numpy as np

from ..utils._compat import Literal
from ..utils._logger import mt
umap_is_installed = False
try:
    from umap import UMAP as cpuUMAP
    umap_is_installed = True
except ImportError:
    pass


cuml_is_installed = False

try:
    import cuml
    from cuml.manifold import UMAP as cumlUMAP
    cuml_is_installed = True
except ImportError:
    pass


def get_default_umap_reducer(**kwargs):
    if 'min_dist' not in kwargs:
        kwargs['min_dist'] = 0.1
    if 'spread' not in kwargs:
        kwargs['spread'] = 1.5
    if 'local_connectivity' not in kwargs:
        kwargs['local_connectivity'] = 1.5
    if 'verbose' not in kwargs:
        kwargs['verbose'] = True
    if 'repulsion_strength' not in kwargs:
        kwargs['repulsion_strength'] = 1.2
    return cpuUMAP(
        **kwargs
    )

def get_constrained_umap_reducer(**kwargs):
    if 'min_dist' not in kwargs:
        kwargs['min_dist'] = 0.1
    if 'spread' not in kwargs:
        kwargs['spread'] = 1.5
    if 'local_connectivity' not in kwargs:
        kwargs['local_connectivity'] = 1.5
    if 'verbose' not in kwargs:
        kwargs['verbose'] = True
    if 'repulsion_strength' not in kwargs:
        kwargs['repulsion_strength'] = 1.2
    return cpuUMAP(
        target_metric='euclidean',
        **kwargs
    )

def get_default_cuml_reducer(**kwargs):
    if 'min_dist' not in kwargs:
        kwargs['min_dist'] = 0.1
    if 'spread' not in kwargs:
        kwargs['spread'] = 1.5
    if 'local_connectivity' not in kwargs:
        kwargs['local_connectivity'] = 1.5
    if 'verbose' not in kwargs:
        kwargs['verbose'] = True
    if 'repulsion_strength' not in kwargs:
        kwargs['repulsion_strength'] = 1.2

    if torch.cuda.is_available():
        return cumlUMAP(
            **kwargs
        )
    else:
        raise Exception('CUDA is not available. Please install CUDA and cuml to use cumlUMAP or use get_default_umap_reducer')

def get_default_reducer():
    if torch.cuda.is_available() and cuml_is_installed:
        return get_default_cuml_reducer()
    else:
        return get_default_umap_reducer()
    

def transfer_umap(
    reference_embedding,
    reference_umap,
    query_embedding,
    method: Literal['retrain','knn'] = 'knn', 
    subsample: int = 100000,
    return_subsampled_indices: bool = False,
    return_subsampled_reference_umap: bool = False,
    n_epochs: int = 10,
    use_cuml_umap: bool = False,
    return_reducer: bool = False,
    **kwargs
):
    """
    Transfer UMAP from reference_embedding to query_embedding

    :param reference_embedding: Reference embedding
    :param reference_umap: Reference UMAP
    :param query_embedding: Query embedding
    :param method: Method to use. Either 'retrain' or 'knn'. 
        If 'retrain', use a new umap reducer. 
        If 'knn', use reference_umap to find nearest neighbors and average their UMAP coordinates
    :param subsample: Number of cells to subsample from reference_embedding.
    :param return_subsampled_indices: Return subsampled indices.
    :param return_subsampled_reference_umap: Return subsampled reference UMAP
    :param n_epochs: Number of epochs to use for retraining. Ignore if method is 'knn'
    :param return_reducer: Return reducer. Ignore if method is 'knn'.
    :param kwargs: Additional arguments to pass to UMAP. Ignore if method is 'knn'.

    :example:
        >>> import scatlasvae
        >>> import scanpy as sc
        >>> import numpy as np
        >>> adata_query.obsm['X_umap'] = scatlasvae.utils.transfer_umap(
        >>>     adata_reference.obsm['X_gex'],
        >>>     adata_reference.obsm['X_umap'],
        >>>     adata_query.obsm['X_gex']
        >>> )["embedding]
    """
    indices = np.arange(len(reference_embedding))
    if subsample < len(indices):
        indices = np.random.choice(indices, size=subsample, replace=False)

    if method == 'retrain':
        mt("Warning: retrain using UMAP reducer is not yet fully developed. Using knn instead.")
        if use_cuml_umap:
            raise NotImplementedError()
        else:
            reducer = get_default_umap_reducer(
                init = reference_umap[indices],
                target_metric = 'euclidean', 
                n_epochs = n_epochs,
                **kwargs
            )

        reducer.fit(reference_embedding[indices], y=reference_umap[indices])
        x = reducer.transform(np.vstack([reference_embedding,query_embedding]))
        z = x[len(reference_embedding):]
        return {
            'embedding': z,
            'reducer': reducer if return_reducer else None,
            'subsampled_indices': indices if return_subsampled_indices else None,
            'subsampled_reference_umap': x[:len(reference_embedding)] if return_subsampled_reference_umap else None
        }

    else:
        from sklearn.neighbors import NearestNeighbors
        knn = NearestNeighbors(n_neighbors=5)
        knn.fit(reference_embedding[indices])
        D, I = knn.kneighbors(query_embedding)
        z = reference_umap[indices][I].mean(1)
        return {
            'embedding': z,
            'reducer': reducer if return_reducer else None,
            'subsampled_indices': indices if return_subsampled_indices else None,
            'subsampled_reference_umap': reference_umap[indices] if return_subsampled_reference_umap else None
        }


from collections import Counter 
import numpy as np
import scanpy as sc 

def cell_type_alignment(
    adata: sc.AnnData, 
    obs_1: str, 
    obs_2: str, 
    palette_1: dict = None,
    palette_2: dict = None,
    perc_in_obs_1: float = 0.1,
    return_fig: bool = True
):
    """
    Plot a Sankey diagram of cell type alignment between two obs columns.

    :param adata: Annotated data matrix.
    :param obs_1: First obs column to compare.
    :param obs_2: Second obs column to compare.
    :param palette_1: Color palette for obs_1.
    :param palette_2: Color palette for obs_2.
    :param perc_in_obs_1: Minimum percentage of cells in obs_1 to be considered.

    :return: Sankey diagram in plotly.graph_objects.Figure format.

    :Example:
        >>> import scatlasvae
        >>> fig = scatlasvae.utils.cell_type_alignment(adata, "cell_type", "cell_type_2")
        >>> fig.show()
        >>> fig.write_html("cell_type_alignment.html")
    """
    if obs_1 not in adata.obs.columns:
        raise ValueError(f"obs_1 {obs_1} not in adata.obs.columns")
    if obs_2 not in adata.obs.columns:
        raise ValueError(f"obs_2 {obs_2} not in adata.obs.columns")
    if palette_1 is None: 
        try:
            palette_1 = sc.pl._tools.scatterplots._get_palette(adata, obs_1)
        except:
            palette_1 = dict(zip(np.unique(adata.obs[obs_1]), ['#000000'] * len(np.unique(adata.obs[obs_1]))))
    if palette_2 is None:
        try:
            palette_2 = sc.pl._tools.scatterplots._get_palette(adata, obs_2)
        except:
            palette_2 = dict(zip(np.unique(adata.obs[obs_2]), ['#000000'] * len(np.unique(adata.obs[obs_2]))))
    count = {}
    c = Counter(adata.obs[obs_1])

    agg = adata.obs.groupby(obs_1).agg({obs_2: Counter})
    for i,j in zip(agg.index,agg.iloc[:,0]):
        for k,v in j.items():
            count[(i,k)]  = v
            
    count = dict(list(filter(lambda x: x[1] / c[x[0][0]] > perc_in_obs_1, count.items())))
    if not return_fig:
        return count
    else:
        try:
            import plotly.graph_objects as go
        except:
            raise ImportError("Please install plotly to use this function.")
        
        labels = list(np.unique(adata.obs[obs_1])) + list(np.unique(adata.obs[obs_2]))
        fig = go.Figure(data=[go.Sankey(
            node = dict(
                pad = 15,
                thickness = 20,
                line = dict(color = "black", width = 0.5),
                label = labels,
                color = "blue"
            ),
            link = dict(
                source = list(map(lambda x: labels.index(x), list(map(lambda z: z[0], count.keys())))), # indices correspond to labels, eg A1, A2, A1, B1, ...
                target = list(map(lambda x: labels.index(x), list(map(lambda z: z[1], count.keys())))),
                value = list(count.values())
            )
        )])

        fig.update_layout(title_text="Basic Sankey Diagram", font_size=10)
        return count, fig